# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import Dict, List, Set, Tuple, Union

import torch
import torch.nn as nn

from brevitas.fx import GraphModule
from brevitas.graph.base import GraphTransform
from brevitas.graph.equalize import _channel_maxabs
from brevitas.graph.equalize import _extract_regions
from brevitas.graph.equalize import _get_input_axis
from brevitas.graph.equalize import _get_output_axis
from brevitas.graph.equalize import Region
from brevitas.graph.equalize import transpose

__all__ = ['GraphChannelSplitting']

_conv = (
    nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)

_unsupported_layers = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.LayerNorm)


def _channels_to_split(
        sources: Dict[str, nn.Module],
        sinks: Dict[str, nn.Module],
        split_criterion: str,
        split_ratio: float,
        split_input: bool) -> Dict[nn.Module, List[torch.Tensor]]:
    """
    This method computes the channels that will be split based on `split_criterion`.
    """
    modules = sinks if split_input else sources
    _get_axis = _get_input_axis if split_input else _get_output_axis
    # the modules are all of the same shape so we can just take the first one
    single_module = next(iter(modules))
    num_channels = single_module.weight.shape[_get_axis(single_module)]
    splits_per_layer = int(math.ceil(split_ratio * num_channels))

    all_channels = []
    if split_criterion == 'maxabs':
        for module in modules:
            # get input/output axis of module
            axis = _get_axis(module)
            # transpose to have axis as first dimension
            weight_t = transpose(module.weight, axis)
            # flatten all but first dimension and get max per channel
            max_per_channel = _channel_maxabs(weight_t.reshape(weight_t.size(0), -1))
            channels_sorted = torch.argsort(max_per_channel, descending=True)
            all_channels.append(channels_sorted[:splits_per_layer])

    # return tensor with the unique indices to split
    channels_to_split = torch.cat(all_channels)
    return torch.unique(channels_to_split)


# decorator is needed to modify the weights in-place using a view
@torch.no_grad()
def _split_channels(
        module: nn.Module,
        channels_to_split: torch.Tensor,
        split_input: bool = False,
        split_factor: float = 0.5) -> None:
    """
    Given a module, this method splits the weight channels as proposed in https://arxiv.org/abs/1901.09504.
    `split_factor` determines how to split the channels, `channels_to_split` is a list of channel indices.
    If `split_input=True`, the input channels of the module are split, otherwise the output channels.
    """
    weight = module.weight.data
    bias = module.bias.data if module.bias is not None else None
    num_added_channels = len(channels_to_split)

    _get_axis = _get_input_axis if split_input else _get_output_axis
    axis = _get_axis(module)
    # save shape of the module weights
    orig_shape = list(weight.shape)
    weight_t = transpose(weight, axis)
    # flatten to 2d
    weight_t = weight_t.reshape(weight_t.size(0), -1)
    for id in channels_to_split:
        # split and get channel to stack
        weight_t[id, :] *= split_factor
        split_channel = weight_t[id, :]
        # expand so we can stack
        split_channel = split_channel.expand(1, split_channel.size(0))
        weight_t = torch.cat([weight_t, split_channel], dim=0)

        if bias is not None and not split_input:
            bias[id] *= split_factor
            split_channel = bias[id:id + 1]
            bias = torch.cat((bias, split_channel))

    # reshape weight_t back to orig shape with the added channels
    del orig_shape[axis]
    weight_t = weight_t.reshape(weight_t.size(0), *orig_shape)
    weight_t = transpose(weight_t, axis)
    module.weight.data = weight_t
    if bias is not None:
        module.bias.data = bias

    if isinstance(module, _conv):
        if split_input:
            module.in_channels += num_added_channels
        else:
            module.out_channels += num_added_channels
    elif isinstance(module, nn.Linear):
        if split_input:
            module.in_features += num_added_channels
        else:
            module.out_features += num_added_channels


def _split_channels_region(
        sources: Dict[str, nn.Module],
        sinks: Dict[str, nn.Module],
        channels_to_split: torch.tensor,
        split_input: bool) -> None:
    if not split_input:
        # splitting output channels
        for module in sources:
            _split_channels(module, channels_to_split, split_input=False)
        for module in sinks:
            # duplicating input_channels for all modules in the sink
            _split_channels(module, channels_to_split, split_factor=1, split_input=True)
    else:
        # input channels are split in half, output channels duplicated
        for module in sinks:
            _split_channels(module, channels_to_split, split_input=True)

        for module in sources:
            # duplicating output_channels for all modules in the source
            _split_channels(module, channels_to_split, split_factor=1, split_input=False)


def _is_groupwise(module: nn.Module) -> bool:
    # only Conv layers can be groupwise
    return isinstance(module, _conv) and module.groups > 1


def _is_unsupported(module: nn.Module) -> bool:
    return isinstance(module, _unsupported_layers)


def _is_mha(module: nn.Module) -> bool:
    return isinstance(module, nn.MultiheadAttention)


def _is_supported(srcs: List[nn.Module], sinks: List[nn.Module]) -> bool:
    # groupwise convolutions are not supported so filter them out
    if any(map(_is_groupwise, srcs + sinks)):
        return False

    # filter out unsupported layers
    if any(map(_is_unsupported, sinks + srcs)):
        return False

    # mha can only be in the sources
    if any(map(_is_mha, sinks)):
        return False
    elif any(map(_is_mha, srcs)):
        # we need to access the weights of the out_proj layers in mha, therefore unwrap
        srcs = _unwrap_mha(srcs)

    # check if OCs of sources are all equal
    srcs_ocs = set(module.weight.shape[_get_output_axis(module)] for module in srcs)
    if len(srcs_ocs) > 1:
        return False

    # check if ICs of sinks are all equal
    sinks_ics = set(module.weight.shape[_get_input_axis(module)] for module in sinks)
    if len(sinks_ics) > 1:
        return False

    return srcs_ocs == sinks_ics


def _unwrap_mha(sources: List[nn.Module]) -> List[nn.Module]:
    for i, source in enumerate(sources):
        if _is_mha(source):
            sources[i] = source.out_proj
    return sources


def _split(
        model: GraphModule,
        regions: List[Region],
        split_ratio: float,
        split_input: bool,
        split_criterion: str = 'maxabs') -> GraphModule:
    for i, region in enumerate(regions):
        sources = [region.get_module_from_name(src) for src in region.srcs_names]
        sinks = [region.get_module_from_name(sink) for sink in region.sinks_names]

        # check for mha in sources and unwrap it for out_proj
        if any(map(_is_mha, sources)):
            sources = _unwrap_mha(sources)

        # get channels to split
        channels_to_split = _channels_to_split(
            sources=sources,
            sinks=sinks,
            split_criterion=split_criterion,
            split_ratio=split_ratio,
            split_input=split_input)
        # splitting/duplicating channels
        _split_channels_region(
            sources=sources,
            sinks=sinks,
            channels_to_split=channels_to_split,
            split_input=split_input)

    return model


def _clean_regions(regions: List[Region]) -> List[Region]:
    """
    Given a list of regions, this method removes all regions that are not compatible with channel splitting.
    """
    # idea: map modules to their regions and check whether it appears in multiple regions
    regions_to_del = set()
    source_modules = dict()
    sink_modules = dict()
    for i, region in enumerate(regions):
        sources = [region.get_module_from_name(src) for src in region.srcs_names]
        sinks = [region.get_module_from_name(sink) for sink in region.sinks_names]

        # a module cannot be in the sources (or sinks) of multiple regions
        for src in sources:
            # if not yet in the dict, instantiate new list for keeping track
            if src not in source_modules:
                source_modules[src] = [i]
            else:
                # we know the module has been in sources before, so region needs to be deleted
                source_modules[src].append(i)
                regions_to_del.update({*source_modules[src]})
        for sink in sinks:
            if sink not in sink_modules:
                sink_modules[sink] = [i]
            else:
                sink_modules[sink].append(i)
                regions_to_del.update({*sink_modules[sink]})

        # check for other unsupported
        if not _is_supported(srcs=sources, sinks=sinks):
            # add region to be deleted
            regions_to_del.add(i)

    regions = [regions[i] for i, _ in enumerate(regions) if i not in regions_to_del]
    return regions


class GraphChannelSplitting(GraphTransform):

    def __init__(
            self,
            split_ratio: float = 0.02,
            split_criterion: str = 'maxabs',
            split_input: bool = True):
        super(GraphChannelSplitting, self).__init__()

        self.split_ratio = split_ratio
        self.split_criterion = split_criterion
        self.split_input = split_input

    def apply(
            self,
            model: GraphModule,
            return_regions: bool = False
    ) -> Union[Tuple[GraphModule, Set[Tuple[str]]], GraphModule]:
        regions = _extract_regions(model)
        regions = _clean_regions(regions)
        if len(regions) > 0:
            model = _split(
                model=model,
                regions=regions,
                split_ratio=self.split_ratio,
                split_criterion=self.split_criterion,
                split_input=self.split_input)
        if return_regions:
            return model, regions
        else:
            return model



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
from inspect import isclass
import operator
from typing import Dict, List, Optional

import torch
import torch.nn as nn

import brevitas
from brevitas.graph.base import InsertModuleCallAfter
from brevitas.graph.base import ModuleInstanceToModuleInstance
from brevitas.graph.base import ModuleToModuleByInstance
from brevitas.graph.utils import del_module
from brevitas.graph.utils import get_module

ADD_FNS = [torch.add, operator.add, operator.iadd]

ADD_METHODS = ['add', 'add_']

SIGN_PRESERVING_MODULES = (
    nn.Dropout,
    nn.Dropout2d,
    nn.Dropout3d,
    nn.MaxPool1d,
    nn.MaxPool2d,
    nn.MaxPool3d,
    nn.AvgPool1d,
    nn.AvgPool2d,
    nn.AvgPool3d,
    nn.AdaptiveAvgPool1d,
    nn.AdaptiveAvgPool2d,
    nn.AdaptiveAvgPool3d,
    nn.PixelShuffle,
    nn.PixelUnshuffle,
    nn.Identity)

PRECISION_PRESERVING_MODULES = (
    nn.Dropout,
    nn.Dropout2d,
    nn.Dropout3d,
    nn.MaxPool1d,
    nn.MaxPool2d,
    nn.MaxPool3d,
    nn.PixelShuffle,
    nn.PixelUnshuffle,
    nn.Identity)

MAX_RESIDUAL_ITERS = 9999

BATCH_NORM = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)


def inp_placeholder_handler(model, input_quantizer):
    """
    Add Quantization step at the input of the network.
    """
    rewriters = []
    if input_quantizer is None:
        return model
    for node in model.graph.nodes:
        if node.op == 'placeholder':
            act_quant, kwargs_act_quant = input_quantizer
            inp_quant = act_quant(**kwargs_act_quant)
            name = node.name + '_quant'
            model.add_module(name, inp_quant)
            rewriters.append(InsertModuleCallAfter(name, node))
    for rewriter in rewriters:
        model = rewriter.apply(model)
    return model


def are_inputs_unsigned(model, node, is_unsigned_list, quant_act_map, unsigned_act_tuple):
    for inp_node in node.all_input_nodes:
        if inp_node.op == 'call_module':
            inp_module = get_module(model, inp_node.target)
            if isinstance(inp_module, tuple(quant_act_map.keys())) and isinstance(
                    inp_module, unsigned_act_tuple):
                is_unsigned_list.append(True)
            elif isinstance(inp_module, tuple(SIGN_PRESERVING_MODULES)):
                are_inputs_unsigned(
                    model, inp_node, is_unsigned_list, quant_act_map, unsigned_act_tuple)
            elif hasattr(inp_module, 'input_quant'):
                is_unsigned_list.append(not inp_module.input_quant.is_signed)
            else:
                is_unsigned_list.append(False)
        elif inp_node.op == 'call_function':
            if inp_node.target in [torch.reshape, torch.flatten, torch.transpose, torch.cat
                                  ] + ADD_FNS:
                are_inputs_unsigned(
                    model, inp_node, is_unsigned_list, quant_act_map, unsigned_act_tuple)
            else:
                is_unsigned_list.append(False)
        elif inp_node.op == 'call_method':
            if inp_node.target in ['view', 'reshape', 'flatten', 't', 'permute'] + ADD_METHODS:
                are_inputs_unsigned(
                    model, inp_node, is_unsigned_list, quant_act_map, unsigned_act_tuple)
            else:
                is_unsigned_list.append(False)
    return all(is_unsigned_list)


def _tensor_quant_in_list(act_quant, module_list, same_sign):
    tq = act_quant.fused_activation_quant_proxy.tensor_quant
    for m in module_list:
        if m is None:
            continue
        m_tq = m.fused_activation_quant_proxy.tensor_quant
        if same_sign and m_tq is tq:
            return True
        elif not same_sign and m_tq.scaling_impl is tq.scaling_impl and m_tq.int_scaling_impl is tq.int_scaling_impl:
            return True
    return False


def are_inputs_quantized_and_aligned(model, node, quantized_modules_list, quant_act_map, same_sign):
    """
    Check if the inputs to `node` are quantized and aligned.
    If same_sign=True, aligned means that the inputs should have same sign and scale factor.
    Otherwise, they need to have only the same scale factors.
    If none of the previous conditions are met (e.g., FP input, or not aligned scales), the function
    returns False.
    """
    for inp_node in node.all_input_nodes:
        if inp_node.op == 'call_module':
            inp_module = get_module(model, inp_node.target)
            if isinstance(inp_module, tuple(PRECISION_PRESERVING_MODULES)) and (
                    not same_sign or
                (same_sign and isinstance(inp_module, tuple(SIGN_PRESERVING_MODULES)))):
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, same_sign)
            elif hasattr(inp_module, 'act_quant'):
                aq = inp_module.act_quant
                if _tensor_quant_in_list(aq, quantized_modules_list, same_sign):
                    continue
                quantized_modules_list.append(aq)
            else:
                quantized_modules_list.append(None)
        elif inp_node.op == 'call_function':
            if inp_node.target in [torch.reshape, torch.flatten, torch.transpose]:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, same_sign)
            elif inp_node.target is torch.cat:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, True)
            elif inp_node.target in ADD_FNS:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, False)
            else:
                quantized_modules_list.append(None)
        elif inp_node.op == 'call_method':
            if inp_node.target in ['view', 'reshape', 'flatten', 't', 'permute']:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, same_sign)
            elif inp_node.target in ADD_METHODS:
                are_inputs_quantized_and_aligned(
                    model, inp_node, quantized_modules_list, quant_act_map, False)
            else:
                quantized_modules_list.append(None)
    if None in quantized_modules_list:
        return False
    elif len(quantized_modules_list) > 1:
        return False
    else:
        return True


def output_quant_handler(
        model,
        node,
        rewriters,
        is_sign_preserving,
        quant_identity_map,
        quant_act_map,
        unsigned_act_tuple=None):
    """
    Starting from `node`, check if any of the users requires requantization (i.e., it does not have
    an act_quant attribute). In that case, the functions adds a requantization step to which all the
    branches are connected. If another branch has its own requantization step, there will be two
    consecutive for that branch.
    """
    if is_sign_preserving and unsigned_act_tuple is None:
        raise RuntimeError("Missing information for output_quant_handler")
    quant_module = None
    quant_module_name = None
    for user in node.users:
        output_quant = True
        if user.op == 'call_module':
            user_module = get_module(model, user.target)
            if hasattr(user_module, 'act_quant'):
                output_quant = False
            elif isinstance(user_module, BATCH_NORM):
                # If the user is BatchNorm, check BN's users and potentially requentize at
                # the output of BN
                output_quant = False
                output_quant_handler(
                    model,
                    user,
                    rewriters,
                    is_sign_preserving,
                    quant_identity_map,
                    quant_act_map,
                    unsigned_act_tuple)
        if output_quant:
            if quant_module_name is None and quant_module is None:
                if is_sign_preserving and are_inputs_unsigned(
                        model, node, [], quant_act_map, unsigned_act_tuple):
                    quant_module_class, quant_module_kwargs = quant_identity_map['unsigned']
                else:
                    quant_module_class, quant_module_kwargs = quant_identity_map['signed']
                quant_module = quant_module_class(**quant_module_kwargs)
                quant_module_name = node.name + '_output_quant'
                model.add_module(quant_module_name, quant_module)
                rewriters.append(InsertModuleCallAfter(quant_module_name, node))


def recursive_input_handler(
        model,
        node,
        shared_quant_identity_name,
        shared_quant_identity,
        rewriters,
        quant_identity_map,
        align_input_quant_fn,
        align_sign):
    """
    For a given CAT or ADD node, iterate through its inputs to make sure they are correctly aligned.
    """
    for inp_node in node.all_input_nodes:
        if inp_node.op == 'call_module':
            module = get_module(model, inp_node.target)
            # Precision preserving modules can be safely traversed
            # In case align_sign is True, the modules should also be sign preserving
            if isinstance(module, tuple(PRECISION_PRESERVING_MODULES)) and (
                    not align_sign or
                (align_sign and isinstance(module, tuple(SIGN_PRESERVING_MODULES)))):
                recursive_input_handler(
                    model,
                    inp_node,
                    shared_quant_identity_name,
                    shared_quant_identity,
                    rewriters,
                    quant_identity_map,
                    align_input_quant_fn,
                    align_sign)
            else:
                # Based on the current module, generate an align_output object
                align_output = align_input_quant_fn(
                    module,
                    shared_quant_identity,
                    shared_quant_identity_name,
                    quant_identity_map,
                    align_sign)
                # If it is a tuple, it is a combination of QuantAct and its configuration that
                # will replace the current inp_node module
                if isinstance(align_output, tuple):
                    quant_module_class, quant_module_kwargs = align_output
                    rewriter = ModuleToModuleByInstance(
                        module, quant_module_class, **quant_module_kwargs)
                    rewriters.append(rewriter)
                # If it is a nn.Module, it is an already instatiated QuantAct that will replace
                # the current inp_node module
                elif isinstance(align_output, torch.nn.Module):
                    rewriter = ModuleInstanceToModuleInstance(module, align_output)
                    rewriters.append(rewriter)
                # If it is a string, we simply add a requantization activation after the current
                # inp_node module
                elif isinstance(align_output, str):
                    rewriters.append(InsertModuleCallAfter(shared_quant_identity_name, inp_node))
                else:
                    assert align_output is None, f"align_output {str(align_output)} not supported."
        elif inp_node.op == 'call_function' and inp_node.target in [
                torch.flatten, torch.reshape, torch.transpose]:
            recursive_input_handler(
                model,
                inp_node,
                shared_quant_identity_name,
                shared_quant_identity,
                rewriters,
                quant_identity_map,
                align_input_quant_fn,
                align_sign)
        elif inp_node.op == 'call_function' and inp_node.target is torch.cat:
            recursive_input_handler(
                model,
                inp_node,
                shared_quant_identity_name,
                shared_quant_identity,
                rewriters,
                quant_identity_map,
                align_input_quant_fn,
                align_sign=True)
        elif inp_node.op == 'call_method' and inp_node.target in [
                'view', 'reshape', 'flatten', 'transpose']:
            recursive_input_handler(
                model,
                inp_node,
                shared_quant_identity_name,
                shared_quant_identity,
                rewriters,
                quant_identity_map,
                align_input_quant_fn,
                align_sign)
        else:
            rewriters.append(InsertModuleCallAfter(shared_quant_identity_name, inp_node))


def _get_quant_module(model, node, quant_identity_map, quant_act_map, unsigned_act_tuple):
    """
    Generate a QuantIdentity node that will be used for requantization step around a node.
    If all inputs to that node are unsigned, the QuantIdentity will also be unsigned, otherwise it
    is signed.
    """
    if are_inputs_unsigned(model, node, [], quant_act_map, unsigned_act_tuple):
        quant_module_class, quant_module_kwargs = quant_identity_map['unsigned']
    else:
        quant_module_class, quant_module_kwargs = quant_identity_map['signed']
    quant_module = quant_module_class(**quant_module_kwargs)
    quant_module_name = node.name + '_quant'
    model.add_module(quant_module_name, quant_module)
    return quant_module, quant_module_name


def residual_handler(
        model, quant_identity_map, quant_act_map, unsigned_act_tuple, align_input_quant_fn):
    iter = 0

    def is_converged(model):

        for node in model.graph.nodes:
            if (node.op == 'call_function' and node.target in ADD_FNS + [torch.cat] or
                    node.op == 'call_method' and node.target in ADD_METHODS):
                rewriters = []
                # If the op is CAT, check that inputs have same sign, and in recursive_input_handler
                # force that the sign is aligned
                same_sign = node.target is torch.cat

                # If input to the CAT or ADD node are quantized and aligned correctly, continue to
                # the next node
                if are_inputs_quantized_and_aligned(model,
                                                    node, [],
                                                    quant_act_map,
                                                    same_sign=same_sign):
                    continue

                # Generate a QuantIdentity module to use for alignement of the inputs
                shared_quant_identity, shared_quant_identity_name = _get_quant_module(
                    model, node, quant_identity_map, quant_act_map, unsigned_act_tuple)

                # Recursively, for every input node, traverse the graph to determine how to quantize
                # and align that input node.
                recursive_input_handler(
                    model,
                    node,
                    shared_quant_identity_name,
                    shared_quant_identity,
                    rewriters,
                    quant_identity_map,
                    align_input_quant_fn,
                    align_sign=same_sign)
                for rewriter in rewriters:
                    model = rewriter.apply(model)
                model.graph.lint()
                model.recompile()
                return False
        return True

    while not is_converged(model):
        iter += 1
        if iter == MAX_RESIDUAL_ITERS:
            raise RuntimeError(
                "Residual handler could not find a solution to align scale factors "
                "across ADDs and CATs")

    return model


def add_output_quant_handler(model, quant_identity_map, quant_act_map, unsigned_act_tuple):
    """
    Check the output of every add node, to decide whether it needs to be requantized or not, based
    on the logic of output_quant_handler.
    """
    rewriters = []
    for node in model.graph.nodes:
        if (node.op == 'call_function' and node.target in ADD_FNS or
                node.op == 'call_method' and node.target in ADD_METHODS):
            output_quant_handler(
                model,
                node,
                rewriters,
                is_sign_preserving=True,
                quant_identity_map=quant_identity_map,
                quant_act_map=quant_act_map,
                unsigned_act_tuple=unsigned_act_tuple)
    for rewriter in rewriters:
        model = rewriter.apply(model)
    return model


def act_handler(model, layer_map):
    for node in model.graph.nodes:
        rewriters = []
        if node.op == 'call_module':
            module = get_module(model, node.target)
            if isinstance(module, tuple(layer_map.keys())):
                if layer_map[type(module)] is not None:
                    quant_module_class, quant_module_kwargs = layer_map[type(module)]
                    quant_module = quant_module_class(**quant_module_kwargs)
                    # Check for activation equalization mul nodes
                    if len(node.users) == 1:
                        user_node = list(node.users.keys())[0]
                        if user_node.name.endswith('act_eq_mul'):
                            # We update activation_impl so that the mul node is executed before quantization
                            act_module = quant_module.act_quant.fused_activation_quant_proxy.activation_impl
                            mul_module = get_module(model, user_node.target)
                            quant_module.act_quant.fused_activation_quant_proxy.activation_impl = torch.nn.Sequential(
                                *[act_module, mul_module])
                            # The mul node added during equalization is removed
                            user_node.replace_all_uses_with(node)
                            model.graph.erase_node(user_node)
                            del_module(model, user_node.target)
                    rewriter = ModuleInstanceToModuleInstance(module, quant_module)
                    rewriters.append(rewriter)
        for rewriter in rewriters:
            model = rewriter.apply(model)
    return model


def layer_handler(
        model,
        layer_map,
        requantize_output,
        quant_identity_map=None,
        quant_act_map=None,
        unsigned_act_tuple=None):
    """
    Replace FP weight layers with their corresponding quantized version
    """
    if requantize_output and (quant_identity_map is None or quant_act_map is None or
                              unsigned_act_tuple is None):
        raise RuntimeError("Missing information to requantize output")
    for node in model.graph.nodes:
        rewriters = []
        if node.op == 'call_module':
            module = get_module(model, node.target)
            if isinstance(module, tuple(layer_map.keys())):
                if requantize_output:
                    if len(node.users) > 1 and all(['getitem' in n.name for n in node.users]):
                        for n in node.users:
                            if len(n.users) > 0:
                                output_quant_handler(
                                    model,
                                    n,
                                    rewriters,
                                    is_sign_preserving=isinstance(module, SIGN_PRESERVING_MODULES),
                                    quant_identity_map=quant_identity_map,
                                    quant_act_map=quant_act_map,
                                    unsigned_act_tuple=unsigned_act_tuple)
                    else:
                        output_quant_handler(
                            model,
                            node,
                            rewriters,
                            is_sign_preserving=isinstance(module, SIGN_PRESERVING_MODULES),
                            quant_identity_map=quant_identity_map,
                            quant_act_map=quant_act_map,
                            unsigned_act_tuple=unsigned_act_tuple)
                if layer_map[type(module)] is not None:
                    quant_module_class, quant_module_kwargs = layer_map[type(module)]
                    # Quantize the input if is not quantized, input_quant is not specified,
                    # and the quant_identity_map is provided.
                    if not are_inputs_quantized_and_aligned(
                            model, node, [], quant_act_map, same_sign=False
                    ) and not 'input_quant' in quant_module_kwargs and len(quant_identity_map) > 0:
                        # Define the source node where to add the requantization step
                        previous_node = node.all_input_nodes[0]
                        # Exclude all the other possible users
                        previous_node_users = list(previous_node.users.keys())
                        previous_node_users.remove(node)

                        act_quant, kwargs_act_quant = quant_identity_map['signed']
                        inp_quant = act_quant(**kwargs_act_quant)
                        name = node.name + '_input_quant'
                        model.add_module(name, inp_quant)
                        rewriter = InsertModuleCallAfter(
                            name, previous_node, tuple(previous_node_users))
                        rewriters.append(rewriter)
                    rewriter = ModuleToModuleByInstance(
                        module, quant_module_class, **quant_module_kwargs)
                    rewriters.append(rewriter)
        for rewriter in rewriters:
            model = rewriter.apply(model)
    return model


def _module_class_name(module_class_or_str):
    name = module_class_or_str.__module__ + '.' + module_class_or_str.__name__ if isclass(
        module_class_or_str) else module_class_or_str
    return name


def find_module(
        model: nn.Module,
        layer_map: Dict[nn.Module, Optional[Dict]],
        module_to_replace: List,
        name_blacklist,
        prefix=''):
    """
    Iterate through the model looking at immediate children of every module to look for supported modules.
    This allows us to stop the search when we meet a top-level module that is supported.
    Specifically, it allows to map nn.MultiheadAttetion to its quantized counterpart and not its
    Linear submodules.
    """
    if _module_class_name(type(model)) in layer_map.keys():
        module_to_replace.append(model)
    else:
        for name, module in model.named_children():
            full_name = prefix + '.' + name if prefix != '' else name
            if name_blacklist is not None and full_name in name_blacklist:
                continue
            find_module(module, layer_map, module_to_replace, name_blacklist, full_name)


def layerwise_layer_handler(
        model: nn.Module, layer_map: Dict[nn.Module, Optional[Dict]], name_blacklist=None):
    """
    Replace FP weight layers with their corresponding quantized version
    """
    # Normalize all module lookups to fully qualified strings
    layer_map = {_module_class_name(m): v for m, v in layer_map.items()}
    module_to_replace = []
    find_module(model, layer_map, module_to_replace, name_blacklist)
    rewriters = []
    for module in module_to_replace:
        if layer_map[_module_class_name(type(module))] is not None:
            quant_module_class, quant_module_kwargs = layer_map[_module_class_name(type(module))]
            rewriter = ModuleToModuleByInstance(module, quant_module_class, **quant_module_kwargs)
            rewriters.append(rewriter)
    for rewriter in rewriters:
        model = rewriter.apply(model)
    return model



# Copyright (C) 2024, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from functools import partial

import numpy as np
import torch
from torch import Tensor
import torch.nn as nn

from brevitas.core.scaling import AccumulatorAwareParameterPreScaling
from brevitas.function.shape import over_output_channels
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL

__all__ = ["apply_bias_correction", "apply_ep_init"]


def get_a2q_module(module: nn.Module):
    for submod in module.modules():
        if isinstance(submod, AccumulatorAwareParameterPreScaling):
            return submod
    return None


def _euclidean_projection_onto_positive_simplex(vec: Tensor, radius: float = 1.):
    assert radius > 0, "Error: radius needs to be strictly positive."
    assert vec.ndim == 1, "Error: projection assumes a vector, not a matrix."
    assert vec.min() >= 0, "Error: assuming a vector of non-negative numbers."
    n_elems = vec.shape[0]
    # if we are already within the simplex, then the best projection is itself
    if vec.sum() <= radius:
        return vec
    # using algorithm derived in `Efficient Projections onto the L1-Ball for
    # Learning in High Dimensions`
    v = vec.cpu().detach().numpy()
    u = np.sort(v)[::-1]
    cumsum_u = np.cumsum(u)
    rho = np.nonzero(u * np.arange(1, n_elems + 1) > (cumsum_u - radius))[0][-1]
    theta = float(cumsum_u[rho] - radius) / (rho + 1)
    w = np.clip(v - theta, 0, np.inf)
    vec.data = torch.tensor(w, dtype=vec.dtype, device=vec.device)
    return vec


def euclidean_projection_onto_l1_ball(vec: Tensor, radius: float):
    assert radius > 0, "Error: radius needs to be strictly positive."
    assert vec.ndim == 1, "Error: projection assumes a vector, not a matrix."
    vec_dir = vec.sign()
    vec_mag = _euclidean_projection_onto_positive_simplex(vec.abs(), radius)
    new_vec = vec_dir * vec_mag
    assert vec.shape == new_vec.shape, "Error: shape changed."
    return new_vec


def l1_proj_matrix_per_channel(weights: Tensor, radius: Tensor):
    assert isinstance(weights, Tensor), "Error: weights is assumed to be a Tensor."
    assert isinstance(radius, Tensor), "Error: radius is assumed to be a Tensor."
    assert weights.ndim == 2, "Error: assuming a matrix with ndim=2."
    # if defined per-tensor
    if radius.ndim == 0:
        radius = torch.ones(weights.shape[0]) * radius
    # if defined per-channel
    else:
        radius = radius.flatten()
        assert radius.nelement() == weights.shape[0], "Error: shape mismatch."
    # project each channel independently
    for i in range(weights.shape[0]):
        w = weights[i]
        z = radius[i].item()
        v = euclidean_projection_onto_l1_ball(w, z)
        weights[i] = v
    return weights


def apply_ep_init(model: nn.Module, inp: Tensor):
    """Euclidean projection-based weight initialization (EP-init) for accumulator-aware
    quantization as proposed in `A2Q+: Improving Accumulator-Aware Weight Quantization`"""
    model.eval()
    dtype = next(model.parameters()).dtype
    device = next(model.parameters()).device

    module_stats = {}
    hook_list = list()

    def register_upper_bound(module: AccumulatorAwareParameterPreScaling, inp, output, name):
        """Accumulate the regularization penalty across constrained layers"""
        nonlocal module_stats

        (weights, input_bit_width, input_is_signed) = inp
        scales: Tensor = module.scaling_impl(weights)
        max_norm: Tensor = module.calc_max_l1_norm(input_bit_width, input_is_signed)

        shape = over_output_channels(weights)
        s = scales.reshape(shape)
        w = weights.reshape(shape)

        z: Tensor = s * max_norm  # radius
        module_stats[name] = (w.detach(), z.detach())  # no gradients

        restrict_value_impl = module.restrict_clamp_scaling.restrict_value_impl
        pre_scaling_init: Tensor = restrict_value_impl.restrict_init_tensor(scales * max_norm)
        assert pre_scaling_init.shape == module.value.shape, "Error: shape mismatch."
        module.value.data = torch.where(
            module.value.data <= pre_scaling_init, module.value.data, pre_scaling_init)

        return output

    # add hooks to each of the A2Q pre-scaling modules
    for name, mod in model.named_modules():
        if isinstance(mod, QuantWBIOL):
            submod = get_a2q_module(mod)
            if submod is not None:
                hook_fn = partial(register_upper_bound, name=name)
                hook = submod.register_forward_hook(hook_fn)
                hook_list.append(hook)

    inp = inp.to(device=device, dtype=dtype)
    model(inp)  # register the scaled upper bounds

    # project weights onto the l1-ball
    for name, mod in model.named_modules():
        if name in module_stats and isinstance(mod, (nn.Conv2d, nn.Linear)):
            (weights, radius) = module_stats[name]
            weights = l1_proj_matrix_per_channel(weights, radius)
            weights = weights.reshape(mod.weight.shape)
            mod.weight.data = weights

    for hook in hook_list:
        hook.remove()

    return model



# Copyright (C) 2024, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
from hashlib import sha256
import os
import random

import numpy as np
import torch
import torch.nn as nn

import brevitas.config as config
from brevitas.export import export_qonnx
import brevitas_examples.imagenet_classification.a2q.utils as utils

parser = argparse.ArgumentParser()
parser.add_argument(
    "--data-root", type=str, required=True, help="Directory where the dataset is stored.")
parser.add_argument(
    "--model-name",
    type=str,
    default="quant_resnet18_w4a4_a2q_32b",
    help="Name of model to train. Default: 'quant_resnet18_w4a4_a2q_32b'",
    choices=utils.model_impl.keys())
parser.add_argument(
    "--save-path",
    type=str,
    default="outputs/",
    help="Directory where to save checkpoints. Default: 'outputs/'")
parser.add_argument(
    "--load-from-path",
    type=str,
    default=None,
    help="Optional local path to load torch checkpoint from. Default: None")
parser.add_argument(
    "--num-workers",
    type=int,
    default=0,
    help="Number of workers for the dataloader to use. Default: 0")
parser.add_argument(
    "--pin-memory",
    action="store_true",
    default=False,
    help="If true, pin memory for the dataloader.")
parser.add_argument(
    "--batch-size", type=int, default=512, help="Batch size for the dataloader. Default: 512")
parser.add_argument(
    "--save-torch-model",
    action="store_true",
    default=False,
    help="If true, save torch model to specified save path.")
parser.add_argument(
    "--export-to-qonnx", action="store_true", default=False, help="If true, export model to QONNX.")

SEED = 0
random.seed(SEED)
np.random.seed(SEED)
torch.manual_seed(SEED)

# create a random input for graph tracing
random_inp = torch.randn(1, 3, 32, 32)

if __name__ == "__main__":

    args = parser.parse_args()

    config.JIT_ENABLED = not args.export_to_qonnx

    # Initialize dataloaders
    print(f"Loading CIFAR10 dataset from {args.data_root}...")
    trainloader, testloader = utils.get_cifar10_dataloaders(
        data_root=args.data_root,
        batch_size_train=args.batch_size,  # does not matter here
        batch_size_test=args.batch_size,
        num_workers=args.num_workers,
        pin_memory=args.pin_memory)

    # if load-from-path is not specified, then use the pre-trained checkpoint
    model = utils.get_model_by_name(args.model_name, pretrained=args.load_from_path is None)
    if args.load_from_path is not None:
        # note that if you used bias correction, you may need to prepare the model for the
        # new biases that were introduced. See `utils.get_model_by_name` for more details.
        state_dict = torch.load(args.load_from_path, map_location="cpu")
        model.load_state_dict(state_dict)
    criterion = nn.CrossEntropyLoss()

    top_1, top_5, loss = utils.evaluate_topk_accuracies(testloader, model, criterion)
    print(f"Final top_1={top_1:.1%}, top_5={top_5:.1%}, loss={loss:.3f}")

    # save checkpoint
    os.makedirs(args.save_path, exist_ok=True)
    if args.save_torch_model:
        ckpt_path = f"{args.save_path}/{args.model_name}.pth"
        torch.save(model.state_dict(), ckpt_path)
        with open(ckpt_path, "rb") as _file:
            bytes = _file.read()
            model_tag = sha256(bytes).hexdigest()[:8]
        new_ckpt_path = f"{args.save_path}/{args.model_name}-{model_tag}.pth"
        os.rename(ckpt_path, new_ckpt_path)
        print(f"Saved model checkpoint to: {new_ckpt_path}")

    if args.export_to_qonnx:
        export_qonnx(
            model.cpu(),
            input_t=random_inp.cpu(),
            export_path=f"{args.save_path}/{args.model_name}-{model_tag}.onnx")



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC

import torch

from brevitas.export.common.handler.base import BaseHandler
from brevitas.export.common.handler.qcdq import CDQCastBiasQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import CDQCastMixin
from brevitas.export.common.handler.qcdq import DQCastMixin
from brevitas.export.common.handler.qcdq import QCDQCastActQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQCastDecoupledWeightQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import \
    QCDQCastDecoupledWeightQuantWithInputProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQCastTruncQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QCDQCastWeightQuantProxyHandlerMixin
from brevitas.export.common.handler.qcdq import QMixin


def _itemize_clip_bounds(clip_args):
    if clip_args is not None:
        clip_args['min_val'] = clip_args['min_val'].item()
        clip_args['max_val'] = clip_args['max_val'].item()
    return clip_args


class TorchDQCastMixin(DQCastMixin, ABC):

    def __init__(self) -> None:
        super().__init__()
        self.symbolic_kwargs = {}

    def dequantize_fn(self, x, scale, zero_point, axis):
        # cast zero_point to float, otherwise if both x
        # and zero_point are uint (as in asym quant)
        # uint - uint can lead to errors. Don't cast x to float
        # as the main float datatype might not be float32 (e.g float16)
        if isinstance(zero_point, torch.Tensor):
            zero_point = zero_point.to(torch.float)
        else:
            zero_point = float(zero_point)
        return (x - zero_point) * scale

    def cast_fn(self, x, dtype):
        return x.type(dtype)

    @property
    def flatten_dequantize_params(self):
        return False

    @property
    def itemize_quantize_scalar_params(self):
        return True

    def validate(self, module):
        assert module.bit_width() > 1., 'Binary quant not supported'


class TorchCDQCastMixin(CDQCastMixin, TorchDQCastMixin, ABC):

    def clip_fn(self, x, min_val, max_val):
        return torch.clamp(x, min_val, max_val)


class TorchQCDQCastMixin(QMixin, TorchCDQCastMixin, ABC):

    @classmethod
    def int8_dtype(cls):
        return torch.qint8

    @classmethod
    def uint8_dtype(cls):
        return torch.quint8

    @classmethod
    def int32_dtype(cls):
        return torch.qint32

    def validate(self, module):
        super().validate(module)
        if getattr(self, '_export_q_node', True):
            assert module.rounding_mode.upper() == 'ROUND', 'Only round to nearest even supported'
        assert not module.is_groupwise, "Export with Per Group quantization not supported"

    def quantize_fn(self, x, scale, zero_point, dtype, axis):
        if axis is None:
            y = torch.quantize_per_tensor(x, scale, zero_point, dtype)
        else:
            y = torch.quantize_per_channel(x, scale, zero_point, axis, dtype)
        return y.int_repr()


class TorchQCDQHandler(BaseHandler):

    def forward(self, *args, **kwargs):
        return self.symbolic_execution(*args, **kwargs)


class TorchQCDQCastWeightQuantProxyHandler(TorchQCDQCastMixin,
                                           QCDQCastWeightQuantProxyHandlerMixin,
                                           TorchQCDQHandler):
    _export_q_node = False

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)


class TorchQCDQCastDecoupledWeightQuantProxyHandler(TorchQCDQCastMixin,
                                                    QCDQCastDecoupledWeightQuantProxyHandlerMixin,
                                                    TorchQCDQHandler):
    _export_q_node = False

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)


class TorchQCDQCastDecoupledWeightQuantWithInputProxyHandler(
        TorchQCDQCastMixin, QCDQCastDecoupledWeightQuantWithInputProxyHandlerMixin,
        TorchQCDQHandler):
    _export_q_node = False

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)


class TorchQCDQCastActQuantProxyHandler(TorchQCDQCastMixin,
                                        QCDQCastActQuantProxyHandlerMixin,
                                        TorchQCDQHandler):

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)


class TorchCDQCastBiasQuantProxyHandler(TorchDQCastMixin,
                                        CDQCastBiasQuantProxyHandlerMixin,
                                        TorchQCDQHandler):
    pass


class TorchQCDQCastTruncQuantProxyHandler(TorchQCDQCastMixin,
                                          QCDQCastTruncQuantProxyHandlerMixin,
                                          TorchQCDQHandler):

    @classmethod
    def int_clip_symbolic_kwargs(cls, narrow, signed, bit_width):
        clip_args = super().int_clip_symbolic_kwargs(narrow, signed, bit_width)
        return _itemize_clip_bounds(clip_args)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import glob
import os
from typing import List, Optional
import warnings

from packaging import version
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution
import torch
from torch import Tensor
from torch.utils import cpp_extension

from brevitas import config
from brevitas import jit as jit

pkg_dir = os.path.dirname(os.path.abspath(__file__))

if torch.__version__.endswith('+cpu'):
    torch_version = version.parse(torch.__version__.rstrip('+cpu'))
else:
    torch_version = version.parse(torch.__version__)

try:
    # Attempt _dynamo import
    is_dynamo_compiling = torch._dynamo.is_compiling
except:
    is_dynamo_compiling = lambda: False

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

if config.JIT_ENABLED or config.NATIVE_STE_BACKEND_ENABLED:
    config.NATIVE_STE_BACKEND_ENABLED = True  # for consistency, in case only JIT_ENABLED was true
    extensions_dir = os.path.join(pkg_dir, 'csrc')
    sources = glob.glob(os.path.join(extensions_dir, '*.cpp'))
    sources = [os.path.join(extensions_dir, s) for s in sources]

    try:
        cpp_extension.load(
            name='autograd_ste_ops',
            sources=sources,
            is_python_module=False,
            verbose=config.VERBOSE)
        NATIVE_STE_BACKEND_LOADED = True
    except Exception as e:
        if config.VERBOSE:
            # Warnings calls str on the message argument, can't pass an f-string directly
            error_message = (
                f"The Brevitas native STE backend is enabled but couldn't be loaded.\n"
                f"Ensure that the \"ninja\" build system is installed (e.g. apt install ninja-build)"
                f"\nException: {e}.")
            warnings.warn(error_message)
        NATIVE_STE_BACKEND_LOADED = False
else:
    NATIVE_STE_BACKEND_LOADED = False

from brevitas import ops as ops



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from brevitas.core.quant import PrescaledRestrictIntQuant
from brevitas.core.quant import RescalingIntQuant
from brevitas.inject import ExtendedInjector
from brevitas.inject import value
from brevitas.inject.enum import QuantType
from brevitas.proxy import BiasQuantProxyFromInjector
from brevitas.quant.solver.common import *
from brevitas.quant.solver.parameter import *
from brevitas.quant.solver.parameter import SolveInputViewImpl

__all__ = [
    'BiasQuantSolver',
    'SolveBiasTensorQuantFromEnum',
    'SolveBiasScalingStatsInputConcatDimFromModule',
    'SolveBiasBitWidthImplFromEnum',
    'SolveBiasScalingPerOutputChannelShapeFromModule']


class SolveBiasScalingStatsInputConcatDimFromModule(ExtendedInjector):
    scaling_stats_input_concat_dim = 0  # bias has only 1 dimension by definition


class SolveBiasScalingPerOutputChannelShapeFromModule(ExtendedInjector):

    @value
    def scaling_per_output_channel_shape(module):
        if isinstance(module, tuple):
            assert all(m.out_channels == module[0].out_channels for m in module)
            module = module[0]
        return (module.out_channels,)


class SolveBiasTensorQuantFromEnum(SolveIntQuantFromEnum):

    @value
    def tensor_quant(quant_type, requires_input_scale):
        if quant_type == QuantType.FP:
            return None
        elif quant_type == QuantType.INT:
            if requires_input_scale:
                return PrescaledRestrictIntQuant
            else:
                return RescalingIntQuant
        elif quant_type == QuantType.TERNARY:
            raise RuntimeError(f'{quant_type} not supported.')
        elif quant_type == QuantType.BINARY:
            raise RuntimeError(f'{quant_type} not supported.')
        else:
            raise RuntimeError(f'{quant_type} not recognized.')


class BiasQuantSolver(SolveScalingStatsInputViewShapeImplFromEnum,
                      SolveParameterScalingShape,
                      SolveStatsReduceDimFromEnum,
                      SolveScalingStatsOpFromEnum,
                      SolveTensorQuantFloatToIntImplFromEnum,
                      SolveRestrictScalingImplFromEnum,
                      SolveIntScalingImplFromEnum,
                      SolveParameterScalingImplFromEnum,
                      SolveParameterTensorClampImplFromEnum,
                      SolveParameterScalingInitFromEnum,
                      SolveBitWidthImplFromEnum,
                      SolveBiasScalingPerOutputChannelShapeFromModule,
                      SolveBiasScalingStatsInputConcatDimFromModule,
                      SolveBiasTensorQuantFromEnum,
                      SolveDtypeDeviceFromTrackedParameterList,
                      SolveInputViewImpl):
    """
    Translate enum directives to bias-specific quantization core modules.
    It should be placed last in the list of classes a quantizer inherits from,
    to make sure overrides are correctly captured.
    """
    proxy_class = BiasQuantProxyFromInjector



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

Adapted from https://github.com/mit-han-lab/llm-awq, released under the following LICENSE:

MIT License

Copyright (c) 2023 MIT HAN Lab

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from collections import defaultdict
import functools
import gc

import torch
import torch.nn as nn
import tqdm
from transformers.models.llama.modeling_llama import LlamaForCausalLM
from transformers.models.opt.modeling_opt import OPTForCausalLM

from .auto_clip import apply_clip
from .auto_clip import auto_clip_block
from .auto_scale import apply_scale
from .auto_scale import auto_scale_block

__all__ = ["run_awq"]


def get_named_linears(module):
    return {name: m for name, m in module.named_modules() if isinstance(m, nn.Linear)}


def get_blocks(model):
    if isinstance(model, LlamaForCausalLM):
        layers = model.model.layers
    elif isinstance(model, OPTForCausalLM):
        layers = model.model.decoder.layers
    else:
        raise NotImplementedError(type(model))
    return layers


@torch.no_grad()
def run_awq(
    model,
    enc,
    w_bit,
    q_config,
    n_samples=512,
    seqlen=512,
    auto_scale=True,
    mse_range=True,
    # some configs for ablation study
    calib_data="pileval",
):
    from .utils.calib_data import get_calib_dataset
    from .utils.module import append_str_prefix
    from .utils.module import get_op_name

    layers = get_blocks(model)

    samples = get_calib_dataset(
        data=calib_data, tokenizer=enc, n_samples=n_samples, block_size=seqlen)
    samples = torch.cat(samples, dim=0)

    inps = []
    layer_kwargs = {}

    # get input and kwargs to layer 0
    # with_kwargs is only supported in PyTorch 2.0
    # use this Catcher hack for now
    class Catcher(nn.Module):

        def __init__(self, module):
            super().__init__()
            self.module = module

        def forward(self, inp, **kwargs):
            inps.append(inp)
            layer_kwargs.update(kwargs)
            raise ValueError  # early exit to break later inference

    # patch layer 0 to catch input and kwargs
    layers[0] = Catcher(layers[0])
    try:
        model(samples.to(next(model.parameters()).device))
    except ValueError:  # work with early exit
        pass
    layers[0] = layers[0].module  # restore
    inps = inps[0]

    gc.collect()
    torch.cuda.empty_cache()

    awq_results = {
        "scale": [],
        "clip": [],}

    # solve layer by layer
    for i in tqdm.tqdm(range(len(layers)), desc="Running AWQ..."):
        layer = layers[i]
        named_linears = get_named_linears(layer)

        # firstly, get input features of all linear layers
        def cache_input_hook(m, x, y, name, feat_dict):
            x = x[0]
            x = x.detach().cpu()
            feat_dict[name].append(x)

        input_feat = defaultdict(list)
        handles = []
        for name in named_linears:
            handles.append(
                named_linears[name].register_forward_hook(
                    functools.partial(cache_input_hook, name=name, feat_dict=input_feat)))
        inps = inps.to(next(layer.parameters()).device)  # in case multi-gpu
        # get output as next layer's input
        inps = layer(inps, **layer_kwargs)[0]
        for h in handles:
            h.remove()

        # now solve for scaling and clipping
        input_feat = {k: torch.cat(v, dim=0) for k, v in input_feat.items()}

        if auto_scale:  # if it applies, we should also modify the input_feat with scales
            scales_list = auto_scale_block(
                layer,
                layer_kwargs,
                w_bit=w_bit,
                q_config=q_config,
                input_feat=input_feat,
            )
            apply_scale(layer, scales_list, input_feat_dict=input_feat)
            # append prefix to make names global
            awq_results["scale"] += append_str_prefix(scales_list, get_op_name(model, layer) + ".")

        if mse_range:
            clip_list = auto_clip_block(
                layer,
                w_bit=w_bit,
                q_config=q_config,
                input_feat=input_feat,
            )
            apply_clip(layer, clip_list)
            # append prefix to make names global
            awq_results["clip"] += append_str_prefix(clip_list, get_op_name(model, layer) + ".")

        del input_feat
        gc.collect()
        torch.cuda.empty_cache()

    return awq_results


def apply_awq(model, awq_results):
    apply_scale(model, awq_results["scale"])
    apply_clip(model, awq_results["clip"])



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch

from brevitas.function.ops import max_int
from brevitas.function.ops import min_int
from brevitas.function.ops_ste import ceil_ste
from brevitas.function.ops_ste import round_ste
from brevitas.quant_tensor import _unpack_quant_tensor
from brevitas.quant_tensor import IntQuantTensorBase
from brevitas.quant_tensor import QuantTensor

from .int_torch_handler import INT_QUANT_TENSOR_FN_HANDLER
from .torch_handler import QUANT_TENSOR_FN_HANDLER

IS_VALID_ATOL = 2e-1
B_FLOAT16_IS_VALID_ATOL = 0.5


class IntQuantTensor(IntQuantTensorBase, QuantTensor):

    def __new__(cls, value, scale, zero_point, bit_width, signed, training):

        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor(scale, dtype=torch.float)
        if not isinstance(zero_point, torch.Tensor):
            zero_point = torch.tensor(zero_point, dtype=torch.float)
        if not isinstance(bit_width, torch.Tensor):
            bit_width = torch.tensor(bit_width, dtype=torch.float)
        if not isinstance(signed, torch.Tensor):
            signed = torch.tensor(signed, dtype=torch.bool)
        if not isinstance(training, torch.Tensor):
            training = torch.tensor(training, dtype=torch.bool)
        quant_tensor = super().__new__(cls, value, scale, zero_point, bit_width, signed, training)
        return quant_tensor

    @property
    def signed(self):
        return self.signed_t.item()

    @property
    def training(self):
        return self.training_t.item()

    def __torch_function__(self, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if func in INT_QUANT_TENSOR_FN_HANDLER:
            return INT_QUANT_TENSOR_FN_HANDLER[func](*args, **kwargs)
        elif func in QUANT_TENSOR_FN_HANDLER:
            return QUANT_TENSOR_FN_HANDLER[func](*args, **kwargs)
        else:
            args = _unpack_quant_tensor(args)
            kwargs = _unpack_quant_tensor(kwargs)
            return func(*args, **kwargs)

    @property
    def tensor(self):
        return self.value

    @property
    def _pre_round_int_value(self):
        value = self.value
        scale = self.scale
        zero_point = self.zero_point
        if self.scale.dtype == torch.bfloat16:
            value = self.value.type(torch.float32)
            scale = self.scale.type(torch.float32)
            zero_point = self.zero_point.type(torch.float32)
        int_value = value / scale
        int_value = int_value + zero_point
        return int_value

    @property
    def is_valid(self):
        with torch.no_grad():
            pre_round_int_value = self._pre_round_int_value
            rounded_int_value = torch.round(pre_round_int_value)
            max_abs_diff = torch.max(torch.abs(pre_round_int_value - rounded_int_value))
            atol = B_FLOAT16_IS_VALID_ATOL if self.value.dtype in (
                torch.bfloat16, torch.float16) else IS_VALID_ATOL
            is_int = max_abs_diff < atol
            if self.bit_width >= 2:
                if self.signed:
                    is_upper_b = (2.0 ** (self.bit_width - 1) - 1 >= rounded_int_value).all()
                    is_lower_b = (-2.0 ** (self.bit_width - 1) <= rounded_int_value).all()
                else:
                    is_upper_b = (2.0 ** self.bit_width - 1 >= rounded_int_value).all()
                    is_lower_b = (0. <= rounded_int_value).all()
                return (is_int & is_upper_b & is_lower_b).item()
            else:  # binary case
                unique_vals = rounded_int_value.unique(
                    sorted=False, return_counts=False, return_inverse=False)
                is_binary = unique_vals.view(-1).size()[0] == 2
                is_signed = (unique_vals < 0.).any().item()
                sign_match = is_signed == self.signed
                return is_int.item() and is_binary and sign_match

    @property
    def device(self):
        value_device = self.value.device
        is_same_device = True
        for t in [self.scale, self.zero_point, self.bit_width]:
            is_same_device &= value_device == t.device
        if not is_same_device:
            raise RuntimeError("Value and metadata are on different devices")
        return value_device

    def int(self, float_datatype=False):
        if self.is_valid:
            int_value = round_ste(self._pre_round_int_value)
            if float_datatype:
                # Values at 8bit and lower can be represented exactly with float16 and bfloat16
                # otherwise (e.g. Int16 bias), we upscale to float32
                if self.bit_width <= 8.:
                    return int_value.type(self.scale.dtype)
                else:
                    return int_value.type(torch.float32)
            else:
                if self.bit_width <= 8. and self.signed_t.item():
                    return int_value.to(torch.int8)
                elif self.bit_width <= 8. and not self.signed_t.item():
                    return int_value.to(torch.uint8)
                else:
                    return int_value.to(torch.int32)
        else:
            raise RuntimeError(f"IntQuantTensor not valid.")

    @staticmethod
    def check_input_type(tensor):
        if not isinstance(tensor, IntQuantTensor):
            raise RuntimeError("Tensor is not a IntQuantTensor")

    @staticmethod
    def is_zero_zero_point(tensor):
        IntQuantTensor.check_input_type(tensor)
        return (tensor.zero_point == 0.).all()

    def check_scaling_factors_same(self, other):
        if self.training:
            return True
        if not torch.allclose(self.scale, other.scale):
            raise RuntimeError("Scaling factors are different")

    def check_zero_points_same(self, other):
        if self.training:
            return True
        if not torch.allclose(self.zero_point, other.zero_point):
            raise RuntimeError("Zero points are different")

    def check_bit_width_same(self, other):
        if not torch.allclose(self.bit_width, other.bit_width):
            raise RuntimeError("Bit widths are different")

    def check_sign_same(self, other):
        if not self.signed == other.signed:
            raise RuntimeError("Signs are different")

    def view(self, *args, **kwargs):
        return self.set(value=self.value.view(*args, **kwargs))

    def reshape(self, *args, **kwargs):
        return self.set(value=self.value.reshape(*args, **kwargs))

    def flatten(self, *args, **kwargs):
        return self.set(value=self.value.flatten(*args, **kwargs))

    def transpose(self, *args, **kwargs):
        value = self.value.transpose(*args, **kwargs)
        tensor_meta = {
            'scale': self.scale, 'zero_point': self.zero_point, 'bit_width': self.bit_width}
        for k, tm in tensor_meta.items():
            if len(value.shape) == len(tm.shape):
                tensor_meta[k] = tm.transpose(*args, **kwargs)
        return self.set(value=value, **tensor_meta)

    def permute(self, *args, **kwargs):
        value = self.value.permute(*args, **kwargs)
        tensor_meta = {
            'scale': self.scale, 'zero_point': self.zero_point, 'bit_width': self.bit_width}
        for k, tm in tensor_meta.items():
            if len(value.shape) == len(tm.shape):
                tensor_meta[k] = tm.permute(*args, **kwargs)
        return self.set(value=value, **tensor_meta)

    def size(self, *args, **kwargs):
        return self.value.size(*args, **kwargs)

    @property
    def ndim(self):
        return self.value.ndim

    def dim(self):
        return self.value.dim()

    @property
    def shape(self):
        return self.value.shape

    def dim(self):
        return self.value.dim()

    def add(self, other):
        return self + other

    @staticmethod
    def cat(tensors, dim, out=None):
        if out is not None:
            raise RuntimeError("Out not supported.")
        if len(tensors) < 2:
            return tensors[0]
        else:
            first_qt = tensors[0]
            if all([isinstance(qt, IntQuantTensor) for qt in tensors]):
                for qt in tensors[1:]:
                    first_qt.check_scaling_factors_same(qt)
                    first_qt.check_zero_points_same(qt)
                    first_qt.check_bit_width_same(qt)
                    first_qt.check_sign_same(qt)
                output_value = torch.cat([qt.value for qt in tensors], dim=dim)
                output_training = any([qt.training for qt in tensors])
                if output_training:
                    output_scale = sum([qt.scale for qt in tensors]) / len(tensors)
                    output_zero_point = sum([qt.zero_point for qt in tensors]) / len(tensors)
                    output_bit_width = sum([qt.bit_width for qt in tensors]) / len(tensors)
                else:  # at eval time, they are the same
                    output_scale = first_qt.scale
                    output_zero_point = first_qt.zero_point
                    output_bit_width = first_qt.bit_width
                output_signed = first_qt.signed  # they are the same
                return IntQuantTensor(
                    value=output_value,
                    scale=output_scale,
                    zero_point=output_zero_point,
                    bit_width=output_bit_width,
                    signed=output_signed,
                    training=output_training)
            else:
                tensors = [_unpack_quant_tensor(qt) for qt in tensors]
                output_value = torch.cat(tensors, dim=dim)
                return output_value

    # Reference: https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types

    def __neg__(self):
        neg_value = (-self.int(float_datatype=True) - self.zero_point) * self.scale
        # In case the dtype of self.int is different from the one of the scale
        neg_value = neg_value.type(self.scale.dtype)
        if self.signed:
            return IntQuantTensor(
                value=neg_value,
                scale=self.scale,
                zero_point=self.zero_point,
                bit_width=self.bit_width,
                signed=self.signed,
                training=self.training)
        else:
            return IntQuantTensor(
                value=neg_value,
                scale=self.scale,
                zero_point=self.zero_point,
                bit_width=self.bit_width + 1,
                signed=True,
                training=self.training)

    def __add__(self, other):
        if isinstance(other, IntQuantTensor):
            self.check_scaling_factors_same(other)
            output_value = self.value + other.value
            output_scale = (self.scale + other.scale) / 2
            output_zero_point = self.zero_point + other.zero_point
            max_val = max_int(signed=self.signed, narrow_range=False, bit_width=self.bit_width)
            max_val += max_int(signed=other.signed, narrow_range=False, bit_width=other.bit_width)
            min_val = min_int(signed=self.signed, narrow_range=False, bit_width=self.bit_width)
            min_val += min_int(signed=other.signed, narrow_range=False, bit_width=other.bit_width)
            output_bit_width = ceil_ste(torch.log2(max_val - min_val))
            output_signed = self.signed or other.signed
            output_training = self.training or other.training
            output = IntQuantTensor(
                value=output_value,
                scale=output_scale,
                zero_point=output_zero_point,
                bit_width=output_bit_width,
                signed=output_signed,
                training=output_training)
        elif isinstance(other, QuantTensor):
            output = self.value + _unpack_quant_tensor(other)
        else:
            # When adding a QT with a normal Tensor, we use the zero_point as a way to preserve
            # and return a QT.
            output = IntQuantTensor(
                value=self.value + _unpack_quant_tensor(other),
                scale=self.scale,
                zero_point=self.zero_point - _unpack_quant_tensor(other) / self.scale,
                bit_width=self.bit_width,
                signed=self.signed,
                training=self.training)
        return output

    def __mul__(self, other):
        if isinstance(other, IntQuantTensor):
            output_value = self.value * other.value
            output_scale = self.scale * other.scale
            output_bit_width = self.bit_width + other.bit_width
            output_signed = self.signed or other.signed
            output_training = self.training or other.training
            if self.is_zero_zero_point(self) and self.is_zero_zero_point(other):
                output_zero_point = self.zero_point * other.zero_point
            else:
                raise RuntimeError("Zero-points of mul operands are non-zero, not supported.")
            output = IntQuantTensor(
                value=output_value,
                scale=output_scale,
                zero_point=output_zero_point,
                bit_width=output_bit_width,
                signed=output_signed,
                training=output_training)
        else:
            output = self.value * _unpack_quant_tensor(other)
        return output

    def __str__(self):
        return f"IntQuantTensor(value={self.value}, scale={self.scale}, zero_point={self.zero_point}, bit_width={self.bit_width}, signed_t={self.signed_t}, training_t={self.training_t})"

    def __truediv__(self, other):
        if isinstance(other, IntQuantTensor):
            output_tensor = self.value / other.value  # Note, output tensor not guaranteed to pass self.is_valid()
            max_int_denominator = 2 ** (other.bit_width - int(other.signed))
            output_scale = self.scale / (other.scale * max_int_denominator)
            output_bit_width = self.bit_width + other.bit_width
            output_signed = self.signed or other.signed
            output_training = self.training or other.training
            if self.is_zero_zero_point(self) and self.is_zero_zero_point(other):
                output_zero_point = self.zero_point * other.zero_point  # Output zero_point is a new, zero-valued tensor
            else:
                raise RuntimeError("Zero-points of div operands are non-zero, not supported.")
            output = IntQuantTensor(
                value=output_tensor,
                scale=output_scale,
                zero_point=output_zero_point,
                bit_width=output_bit_width,
                signed=output_signed,
                training=output_training)
        else:
            output = self.value / _unpack_quant_tensor(other)
        return output

    def __abs__(self):
        if self.signed:
            abs_value = (torch.abs(self.int(float_datatype=True)) - self.zero_point) * self.scale
            # In case the dtype of self.int is different from the one of the scale
            abs_value = abs_value.type(self.scale.dtype)
            return IntQuantTensor(
                value=abs_value,
                scale=self.scale,
                zero_point=self.zero_point,
                bit_width=self.bit_width - 1,
                signed=False,
                training=self.training)
        else:
            return self

    def __pos__(self):
        return self



import warnings

import torch
from transformers.models.opt.modeling_opt import OPTAttention

from brevitas.graph import ModuleToModuleByClass
from brevitas_examples.llm.llm_quant.mha_layers import QuantizableOPTAttention

QUANTIZABLE_MHA_MAP = {OPTAttention: (QuantizableOPTAttention, {'batch_first': True})}


def replace_mha_with_quantizable_layers(model, dtype):
    rewriters = []
    for src_module, (quantizable_module, quantizable_module_kwargs) in QUANTIZABLE_MHA_MAP.items():
        rewriter = ModuleToModuleByClass(
            src_module, quantizable_module, **quantizable_module_kwargs, dtype=dtype)
        rewriters.append(rewriter)
    if not rewriters:
        warnings.warn(
            f"No module to replace was found. Supported modules are {list(QUANTIZABLE_MHA_MAP.keys())}"
        )
    for rewriter in rewriters:
        model = rewriter.apply(model)
    return model


@torch.no_grad()
def add_zero_bias_to_linear(model: torch.nn.Module) -> torch.nn.Module:
    for name, module in model.named_modules():
        if type(module) == torch.nn.Linear:
            if module.bias is None:
                module.register_parameter(
                    "bias",
                    torch.nn.Parameter(
                        torch.zeros((module.weight.shape[0],),
                                    device=module.weight.device,
                                    dtype=module.weight.dtype)),
                )
    return model



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from configparser import ConfigParser
import os
from typing import Dict, Optional
from warnings import warn

from torch import hub

from .mobilenetv1 import *
from .proxylessnas import *
from .vgg import *

model_impl = {
    'quant_mobilenet_v1': quant_mobilenet_v1,
    'quant_proxylessnas_mobile14': quant_proxylessnas_mobile14}


def model_with_cfg(name: str, pretrained: Optional[str] = None, bit_width: Optional[int] = None):
    cfg = ConfigParser()
    current_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(current_dir, './cfg', name + '.ini')
    assert os.path.exists(config_path)
    cfg.read(config_path)
    arch = cfg.get('MODEL', 'ARCH')

    if bit_width is not None:
        cfg.set('QUANT', 'BIT_WIDTH', str(bit_width))

    model = model_impl[arch](cfg)
    if pretrained is not None:
        if pretrained == 'quant_weights':
            weight_url = 'PRETRAINED_URL'
        elif pretrained == 'float_weights':
            weight_url = 'FLOAT_PRETRAINED_URL'
        else:
            raise RuntimeError("Pretrained setting not supported")

        checkpoint = cfg.get('MODEL', weight_url)
        state_dict = hub.load_state_dict_from_url(checkpoint, progress=True, map_location='cpu')
        model.load_state_dict(state_dict, strict=True)
    return model, cfg


def quant_mobilenet_v1(float_pretrained: bool = True, bit_width: int = 8):
    pretrained = 'float_weights' if float_pretrained else None
    model, _ = model_with_cfg('quant_mobilenet_v1', pretrained=pretrained, bit_width=bit_width)
    return model


def quant_mobilenet_v1_4b(pretrained: bool = True):
    pretrained = 'quant_weights' if pretrained else None
    model, _ = model_with_cfg('quant_mobilenet_v1_4b', pretrained)
    return model


def quant_mobilenet_v1_4b_round_avgpool(pretrained: bool = False):
    if pretrained:
        warn(
            "The model was trained with floor TruncAvgPool rather than round,"
            " accuracy will be affected.")
    pretrained = 'quant_weights' if pretrained else None
    model, _ = model_with_cfg('quant_mobilenet_v1_4b_round_avgpool', pretrained)
    return model


def quant_proxylessnas_mobile14_4b(pretrained: bool = True):
    pretrained = 'quant_weights' if pretrained else None
    model, _ = model_with_cfg('quant_proxylessnas_mobile14_4b', pretrained)
    return model


def quant_proxylessnas_mobile14_4b_round_avgpool(pretrained: bool = False):
    if pretrained:
        warn(
            "The model was trained with floor TruncAvgPool rather than round,"
            " accuracy will be affected.")
    pretrained = 'quant_weights' if pretrained else None
    model, _ = model_with_cfg('quant_proxylessnas_mobile14_4b_round_avgpool', pretrained)
    return model


def quant_proxylessnas_mobile14_4b5b(pretrained: bool = True):
    pretrained = 'quant_weights' if pretrained else None
    model, _ = model_with_cfg('quant_proxylessnas_mobile14_4b5b', pretrained)
    return model


def quant_proxylessnas_mobile14_hadamard_4b(pretrained: bool = True):
    pretrained = 'quant_weights' if pretrained else None
    model, _ = model_with_cfg('quant_proxylessnas_mobile14_hadamard_4b', pretrained)
    return model



"""
Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
Copyright (c) 2011-2013 NYU                      (Clement Farabet)
Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
   NEC Laboratories America and IDIAP Research Institute nor the names
   of its contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

Forked from PyTorch 1.11.0 as is.
"""

import torch._C

import contextlib
import ctypes
import sys
import types

import torch.jit
import torch._utils_internal

# Query `hasattr` only once.
_SET_GLOBAL_FLAGS = hasattr(sys, 'getdlopenflags') and hasattr(sys, 'setdlopenflags')


@contextlib.contextmanager
def dl_open_guard():
    """
    Context manager to set the RTLD_GLOBAL dynamic linker flag while we open a
    shared library to load custom operators.
    """
    if _SET_GLOBAL_FLAGS:
        old_flags = sys.getdlopenflags()
        sys.setdlopenflags(old_flags | ctypes.RTLD_GLOBAL)
    yield
    if _SET_GLOBAL_FLAGS:
        sys.setdlopenflags(old_flags)

# Each OpOverload object contains pointer to a a specific operator overload, a pointer to the parent `OpOverloadPacket` object.
# You can obtain an OpOverload object through attribute query on OpOverloadPacket.
class OpOverload:
    def __init__(self, overloadpacket, op, schema):
        self._op = op
        self._schema = schema
        self._overloadpacket = overloadpacket

    # it's a no-op since OpOverload object is immutable and must be unique for a given op overload.
    def __deepcopy__(self, memo=None):
        return self

    def __str__(self):
        return "OpOverload(op='{}.{}', overload='{}')".format(*self._schema.name.split("::"), self.overload_name)

    def __call__(self, *args, **kwargs):
        return self._op(*args, **kwargs or {})

    def __getattr__(self, key):
        return getattr(self._op, key)

    # `my_namespace::my_op`
    @property
    def name(self):
        return "{}.{}".format(*self._schema.name.split("::"))

    @property
    def overload_name(self):
        return self._schema.overload_name

    @property
    def overload_packet(self):
        return self._overloadpacket

    @property
    def op(self):
        return self._op

    # TODO: add more methods to expose information about input and output arguments

# OpOverloadPacket class contains pointer to a base unresolved operator that doesn't correspond to a specific operator
# You can obtain an OpOverload object through attribute query.
class OpOverloadPacket:
    def __init__(self, qualified_op_name, op_name, op):
        # These attributes are accessible on the object through the properties
        # defined below but are immutable
        self._qualified_op_name = qualified_op_name
        self._op_name = op_name
        self._op = op

    # it's a no-op since OpOverloadPacket object is immutable and must be unique for a given op.
    def __deepcopy__(self, memo=None):
        return self

    def __str__(self):
        return "OpOverloadPacket(op='{}.{}')".format(*self._qualified_op_name.split("::"))

    @property
    def qualified_op_name(self):
        return "{}.{}".format(*self._qualified_op_name.split("::"))

    @property
    def op_name(self):
        return self._op_name

    @property
    def op(self):
        return self._op

    def __getattr__(self, key):
        # It is not a valid op_name when __file__ is passed in
        if key == '__file__':
            return 'torch.ops'

        try:
            use_key = '' if key == 'default' else key
            # TODO: disallow access to overloads registered by JIT
            op_ = torch._C._get_operation_overload(self._qualified_op_name, use_key)
            schema = torch._C._get_schema(self._qualified_op_name, use_key)
            overload = OpOverload(self, op_, schema)
            # cache the overload object
            setattr(self, key, overload)
            return overload
        except RuntimeError:
            try:
                # This is added to maintain bc in case the user queries an attribute that exists on `self._op`
                # which used to be returned before instead of the OpOverloadPacket
                out = getattr(self._op, key)
                return out
            except AttributeError:
                raise AttributeError("'{}' object has no attribute '{}'".format(str(self), key)) from None

    def __call__(self, *args, **kwargs):
        # overloading __call__ to ensure torch.ops.foo.bar() is still callable from JIT
        # We save the function ptr as the `op` attribute on OpOverloadPacket to access it here.
        return self._op(*args, **kwargs or {})

# Resolution of torch.fn is different from torch.ops.aten.fn
# torch.fn uses the Python argparser, matches with the appropriate schema, and calls into the unboxed version of the method
# torch.ops.aten.fn resolution is done via the mechanism defined in JIT. JIT creates a stack of all the overloads and
# then tries to match the correct one at runtime and always calls into the boxed version of the method
# Autograd codegen creates VariableType, TracerType, inplace or view type and python bindings
# Aten codegen generates tensor methods for the the tensor class

# _OpNamespace is a subclass of ModuleType because the torch script
# allows attribute lookups on modules only. Since we want torch.ops.foo.bar()
# to work from script, we need to ensure ops and foo are modules
class _OpNamespace(types.ModuleType):
    """
    An op namespace to dynamically bind Operators into Python.

    Say a user has created a custom Operator called "my_namespace::my_op". To
    call this op, the user will write torch.ops.my_namespace.my_op(...).
    At startup, this operation will not yet be bound into Python. Instead, the
    following sequence of magic tricks will occur:
    1. `torch.ops.my_namespace` will invoke the `__getattr__` magic method
       on the `torch.ops` object, which will create a new `_OpNamespace`
       object called `my_namespace` and set it as an attribute on the `ops`
       object.
    2. `torch.ops.my_namespace.my_op` will then invoke `__getattr__` on
       the `my_namespace` object, which will retrieve the operation via
       `torch.get_operation`, a function bound from C++, and then in a similar
       fashion bind this new object onto the `my_namespace` object.
    3. `torch.ops.my_namespace.my_op(...)` then calls this new operation
        and subsequent accesses will incur no further lookup (the namespace and
        operation will already exist).
    """
    def __init__(self, name):
        super(_OpNamespace, self).__init__('torch.ops.' + name)
        self.name = name

    def __getattr__(self, op_name):
        # It is not a valid op_name when __file__ is passed in
        if op_name == '__file__':
            return 'torch.ops'
        # Get the op `my_namespace::my_op` if available. This will also check
        # for overloads and raise an exception if there are more than one.
        namespace_name = self.name
        qualified_op_name = '{}::{}'.format(namespace_name, op_name)
        op = torch._C._jit_get_operation(qualified_op_name)

        # let the script frontend know that op is identical to the builtin op
        # with qualified_op_name
        torch.jit._builtins._register_builtin(op, qualified_op_name)
        op.__module__ = self.__module__ + "." + namespace_name
        # opoverloadpacket = OpOverloadPacket(qualified_op_name, op_name, op)
        # opoverloadpacket.__module__ = self.__module__ + "." + namespace_name
        # cache the opoverloadpacket to ensure that each op corresponds to
        # a unique OpOverloadPacket object
        # setattr(self, op_name, opoverloadpacket)
        setattr(self, op_name, op)
        return op

class _Ops(types.ModuleType):
    __file__ = '_ops.py'

    def __init__(self):
        super(_Ops, self).__init__('torch.ops')
        self.loaded_libraries = set()

    def __getattr__(self, name):
        # Here we are creating `torch.ops.my_namespace`
        namespace = _OpNamespace(name)
        setattr(self, name, namespace)
        return namespace

    def load_library(self, path):
        """
        Loads a shared library from the given path into the current process.

        The library being loaded may run global initialization code to register
        custom operators with the PyTorch JIT runtime. This allows dynamically
        loading custom operators. For this, you should compile your operator
        and the static registration code into a shared library object, and then
        call ``torch.ops.load_library('path/to/libcustom.so')`` to load the
        shared object.

        After the library is loaded, it is added to the
        ``torch.ops.loaded_libraries`` attribute, a set that may be inspected
        for the paths of all libraries loaded using this function.

        Args:
            path (str): A path to a shared library to load.
        """
        if sys.executable == "torch_deploy":
            return

        path = torch._utils_internal.resolve_library_path(path)
        with dl_open_guard():
            # Import the shared library into the process, thus running its
            # static (global) initialization code in order to register custom
            # operators with the JIT.
            ctypes.CDLL(path)
        self.loaded_libraries.add(path)

# The ops "namespace"
ops = _Ops()


# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .weighted_bit_width import ActivationBitWidthWeightedBySize
from .weighted_bit_width import MEGA
from .weighted_bit_width import QuantLayerOutputBitWidthWeightedByOps
from .weighted_bit_width import WeightBitWidthWeightedBySize



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from copy import deepcopy
from functools import partial
import sys

import torch
from torch import nn
import torch.nn.functional as F

from brevitas.nn import QuantHardTanh
from brevitas.nn import QuantLinear
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from brevitas.proxy.parameter_quant import BiasQuantProxyFromInjectorBase
from brevitas.proxy.parameter_quant import WeightQuantProxyFromInjectorBase
from brevitas.proxy.runtime_quant import ActQuantProxyFromInjectorBase
from brevitas.proxy.runtime_quant import ClampQuantProxyFromInjector
from brevitas.proxy.runtime_quant import TruncQuantProxyFromInjector
from brevitas.quant_tensor import QuantTensor

from .base import Transform

__all__ = [
    'ClipFloatWeights',
    'DisableEnableQuantization',
    'bias_correction_mode',
    'calibration_mode',
    'load_quant_model_mode']

_PARAM_PROXIES = (WeightQuantProxyFromInjectorBase, BiasQuantProxyFromInjectorBase)

_BIAS_PROXIES = (BiasQuantProxyFromInjectorBase)

_ACC_PROXIES = (TruncQuantProxyFromInjector, ClampQuantProxyFromInjector)

_LAYERS_TO_CLIP = (
    nn.Conv1d,
    nn.Conv2d,
    nn.Conv3d,
    nn.Linear,
    nn.ConvTranspose1d,
    nn.ConvTranspose2d,
    nn.ConvTranspose3d)

BN_LAYERS = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)


def disable_return_quant_tensor(model):
    previous_state = {}
    for module in model.modules():
        if hasattr(module, 'return_quant_tensor'):
            previous_state[module] = module.return_quant_tensor
            module.return_quant_tensor = False
    return previous_state


def restore_return_quant_tensor(model, previous_state):
    for module in model.modules():
        if hasattr(module, 'return_quant_tensor') and module in previous_state:
            module.return_quant_tensor = previous_state[module]


def extend_collect_stats_steps(module):
    if hasattr(module, 'collect_stats_steps'):
        # We extend the collect steps in PTQ to match potentially long calibrations
        module.collect_stats_steps = sys.maxsize


def set_collect_stats_to_average(module):
    if hasattr(module, 'collect_stats_steps') and hasattr(module, 'momentum'):
        # Set the average for stats collection to a true average (default is EMA)
        module.momentum = None


def finalize_collect_stats(module):
    if hasattr(module, 'collect_stats_steps') and hasattr(module, 'counter'):
        # If the counter has already reached collect_stats_steps, we do not want to reset it
        # otherwise the restrict_preprocess might be applied twice: during calibration
        # (that happens in training mode) and then when the model is evaluated
        module.counter = max(module.collect_stats_steps, module.counter)


class calibration_mode:

    def __init__(self, model, enabled=True):
        self.model = model
        self.previous_training_state = model.training
        self.disable_quant_inference = DisableEnableQuantization(call_act_quantizer_impl=True)
        self.enabled = enabled
        self.return_quant_tensor_state = dict()

    def __enter__(self):
        if self.enabled:
            self.model.apply(extend_collect_stats_steps)
            self.model.apply(set_collect_stats_to_average)
            self.return_quant_tensor_state = disable_return_quant_tensor(self.model)
            self.disable_quant_inference.apply(
                self.model, is_training=True, quantization_enabled=False)

    def __exit__(self, type, value, traceback):
        self.model.apply(finalize_collect_stats)
        self.disable_quant_inference.apply(
            self.model, is_training=self.previous_training_state, quantization_enabled=True)
        restore_return_quant_tensor(self.model, self.return_quant_tensor_state)


class bias_correction_mode:

    def __init__(self, model, enabled=True, skip_if_no_bias=False):
        self.model = model
        self.bias_correction = _BiasCorrection(skip_if_no_bias=skip_if_no_bias)
        self.enabled = enabled
        self.hooks = []
        self.output_quant_modules = []

    def __enter__(self):
        if self.enabled:
            for module in self.model.modules():
                # Disable output quant so that the bias correction can be merged in the bias
                if hasattr(module, 'output_quant') and module.output_quant.is_quant_enabled:
                    self.bias_correction.disable_act_quantization(
                        module.output_quant, is_training=False)
                    self.output_quant_modules.append(module)
            self.bias_correction.register_hook_to_wbiol(self.model, self.hooks)

    def __exit__(self, type, value, traceback):
        self.bias_correction.apply_correction(self.model)
        for module in self.output_quant_modules:
            # Re-enable output quantization
            self.bias_correction.enable_act_quantization(module.output_quant, is_training=False)
        for hook in self.hooks:
            hook.remove()


class load_quant_model_mode:

    def __init__(self, model):
        self.model = model
        self.tracked_modules = []

    def __enter__(self):
        for module in self.model.modules():
            if issubclass(type(module), QuantWBIOL):
                module._quant_load_model_mode = True

    def __exit__(self, *args, **kwargs):
        for module in self.model.modules():
            if issubclass(type(module), QuantWBIOL):
                module._quant_load_model_mode = False


class ClipFloatWeights(Transform):

    def __init__(self, threshold=15., layers_to_clip=_LAYERS_TO_CLIP) -> None:
        super(ClipFloatWeights, self).__init__()
        self.threshold = threshold
        self.layers_to_clip = layers_to_clip

    def apply(self, model):
        for module in model.modules():
            if isinstance(module, self.layers_to_clip):
                module.weight.data.clamp_(-self.threshold, self.threshold)
        return model


class DisableEnableQuantization(Transform):

    def __init__(self, call_act_quantizer_impl=False):
        super(DisableEnableQuantization, self).__init__()
        self.disable_act_quant_hooks = []
        self.call_act_quantizer_impl = call_act_quantizer_impl

    def unpack_input(self, inp):
        if isinstance(inp, tuple):
            inp = inp[0]
        if isinstance(inp, QuantTensor):
            inp = inp.value
        return inp

    def disable_act_quant_hook(self, module, inp, output):
        inp = self.unpack_input(inp)
        if module.fused_activation_quant_proxy is not None:
            inp = module.fused_activation_quant_proxy.activation_impl(inp)
        # consider the first module as representative for the activation fn
        # as this is what would happen with a shared act_quant
        # this gets called both during (empty) input_quant and act_quant
        # but for HardTanh it's not an issue
        if isinstance(module.tracked_module_list[0], QuantHardTanh):
            inp = F.hardtanh(
                inp, min_val=module.quant_injector.min_val, max_val=module.quant_injector.max_val)
        return inp

    def disable_act_quantization(self, model, is_training):
        # If self.call_act_quantizer_impl is set to True, the quantization will be performed but the output
        # will be discarded through the hook. It is useful for collecting activation stats,
        # for example during activation calibration in PTQ
        for module in model.modules():
            if isinstance(module, ActQuantProxyFromInjectorBase):
                module.train(is_training)
                if self.call_act_quantizer_impl:
                    for m in module.modules():
                        if hasattr(m, 'observer_only'):
                            m.observer_only = True
                else:
                    module.disable_quant = True
            elif isinstance(module, _ACC_PROXIES):
                module.train(is_training)
                module.disable_quant = True

    def disable_param_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _PARAM_PROXIES):
                module.train(is_training)
                module.disable_quant = True

    def disable_bias_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _BIAS_PROXIES):
                module.train(is_training)
                module.disable_quant = True

    def enable_act_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _ACC_PROXIES):
                module.train(is_training)
                module.disable_quant = False
            elif isinstance(module, ActQuantProxyFromInjectorBase):
                module.disable_quant = False
                module.train(is_training)
                for m in module.modules():
                    if hasattr(m, 'observer_only'):
                        m.observer_only = False

    def enable_param_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _PARAM_PROXIES):
                module.disable_quant = False
                module.train(is_training)

    def enable_bias_quantization(self, model, is_training):
        for module in model.modules():
            if isinstance(module, _BIAS_PROXIES):
                module.disable_quant = False
                module.train(is_training)

    def apply(self, model, is_training, quantization_enabled):
        if not quantization_enabled:
            self.disable_act_quantization(model, is_training)
            self.disable_param_quantization(model, is_training)
        else:
            self.enable_act_quantization(model, is_training)
            self.enable_param_quantization(model, is_training)


class _BiasCorrection(DisableEnableQuantization):

    LAYERS = (QuantWBIOL,)

    def __init__(self, layers=LAYERS, skip_if_no_bias=False):
        super(_BiasCorrection, self).__init__()
        self.layers = layers
        self.iterations = {}
        self.correction_map = {}
        self.float_mean_map = {}
        self.collect_float_mean_hooks = []
        self.correct_bias_hooks = []
        self.skip_if_no_bias = skip_if_no_bias

    def compute_mean(self, inp, transpose_dim):
        inp = inp.transpose(0, transpose_dim)
        return inp.reshape(inp.shape[0], -1).mean(dim=1).detach()

    def channel_dim(self, inp, module):
        if len(inp.shape) == 3 and isinstance(module, QuantLinear):
            channel_dim = 2
        else:
            channel_dim = 1
        return channel_dim

    def collect_float_mean(self, module, inp, name):
        inp = self.unpack_input(inp)
        if name in self.float_mean_map.keys():
            raise RuntimeError("Module to bias-correct called multiple times, not supported.")
        transpose_dim = self.channel_dim(inp, module)
        self.float_mean_map[name] = self.compute_mean(inp, transpose_dim)

    def update_correction(self, name, error):
        if name not in self.correction_map:
            self.correction_map[name] = error
        else:
            self.correction_map[name] += error

    def apply_correction(self, model):
        for name, module in model.named_modules():
            if name in self.correction_map.keys():
                correction = self.correction_map[name] / self.iterations[name]
                # When accelerate is enabled, bring tensors onto the device to avoid allocating a meta parameter.
                if hasattr(module, 'allocate_params'):
                    module.allocate_params(module)
                if module.bias is not None:
                    module.bias.data += correction
                elif self.skip_if_no_bias is False:
                    # If accelerate is enabled, bias will be on the same execution device as the weights, but won't be managed properly by accelerate
                    module.register_parameter(
                        'bias', nn.Parameter(correction).to(module.weight.device))
                # Offload params again
                if hasattr(module, 'offload_params'):
                    module.offload_params(module)

    def compute_correct_bias(self, module, inp, name):
        inp = self.unpack_input(inp)
        if name in self.float_mean_map.keys():
            transpose_dim = self.channel_dim(inp, module)
            quant_mean = self.compute_mean(inp, transpose_dim)
            error = self.float_mean_map[name] - quant_mean
            self.update_correction(name, error)
            del self.float_mean_map[name]

    def register_hook_to_wbiol(self, model, hooks):
        """
        Forward hooks are registered to the WBIOL layers.
        In this way, if more hooks are registered to these layers, they will be only called once per module call.
        We perform two more forwards for each WBIOL, but since we call forward and not __call__, eventual hooks would be skipped.
        This is a desired behaviour, since these extra forwards are only necessary to compute the bias correction factor.

        If we registered a single hook for the entire model, and hooks were added to WBIOL layers,
        these would be called by our extra forwards, which would be an unexpected behaviours.
        """
        for name, module in model.named_modules():
            if isinstance(module, self.layers):
                self.iterations[name] = 0
                hook_fn = partial(self.forward_hook_wbiol, name=name)
                hooks.append(module.register_forward_hook(hook_fn))

    def forward_hook_wbiol(self, module, inp, output, name):
        """
        After each forward, we perform two extra forwards to register float and quant outputs
        and the error between the two.

        We do not return the original quant output, but the float one, to avoid error accumulation
        """
        # Compute float reference
        self.disable_act_quantization(module, is_training=False)
        self.disable_param_quantization(module, is_training=False)

        out_float = module.forward(*inp)  # Required to avoid infinite recursion
        self.collect_float_mean(module, out_float, name)
        self.enable_act_quantization(module, is_training=False)
        self.enable_param_quantization(module, is_training=False)
        # Keep output quant disabled until further notice
        self.disable_act_quantization(module.output_quant, is_training=False)
        out_quant = output
        self.compute_correct_bias(module, out_quant, name)
        self.iterations[name] += 1
        return out_float


class norm_correction_mode:

    def __init__(self, model, enabled=True):
        self.model = model
        self.enabled = enabled

    def __enter__(self):
        if self.enabled:
            for m in self.model.modules():
                if isinstance(m, BN_LAYERS):
                    m.register_buffer("old_running_mean", m.running_mean.clone())
                    m.register_buffer("old_running_var", m.running_var.clone())
                    m.reset_running_stats()
                    m.momentum = None
                    m.train()

    def __exit__(self, type, value, traceback):
        if self.enabled:
            for m in self.model.modules():
                if isinstance(m, BN_LAYERS):
                    a = torch.sqrt(m.old_running_var / m.running_var)
                    b = m.old_running_mean - a * m.running_mean
                    m.running_var = m.old_running_var.clone() / a
                    m.running_mean = (m.old_running_mean.clone() - b) / a
                    del m.old_running_var
                    del m.old_running_mean



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from functools import partial
from typing import Union

from torch import hub
import torch.nn as nn

from .common import CommonIntAccumulatorAwareZeroCenterWeightQuant
from .espcn import *

model_impl = {
    'float_espcn_x2':
        partial(float_espcn, upscale_factor=2),
    'quant_espcn_x2_w8a8_base':
        partial(quant_espcn_base, upscale_factor=2, weight_bit_width=8, act_bit_width=8),
    'quant_espcn_x2_w8a8_a2q_32b':
        partial(
            quant_espcn_a2q,
            upscale_factor=2,
            weight_bit_width=8,
            act_bit_width=8,
            acc_bit_width=32),
    'quant_espcn_x2_w8a8_a2q_16b':
        partial(
            quant_espcn_a2q,
            upscale_factor=2,
            weight_bit_width=8,
            act_bit_width=8,
            acc_bit_width=16),
    'quant_espcn_x2_w4a4_base':
        partial(quant_espcn_base, upscale_factor=2, weight_bit_width=4, act_bit_width=4),
    'quant_espcn_x2_w4a4_a2q_32b':
        partial(
            quant_espcn_a2q,
            upscale_factor=2,
            weight_bit_width=4,
            act_bit_width=4,
            acc_bit_width=32),
    'quant_espcn_x2_w4a4_a2q_13b':
        partial(
            quant_espcn_a2q,
            upscale_factor=2,
            weight_bit_width=4,
            act_bit_width=4,
            acc_bit_width=13),
    'quant_espcn_x2_w4a4_a2q_plus_13b':
        partial(
            quant_espcn,
            upscale_factor=2,
            weight_bit_width=4,
            act_bit_width=4,
            acc_bit_width=13,
            weight_quant=CommonIntAccumulatorAwareZeroCenterWeightQuant),
    'quant_espcn_x2_w8a8_a2q_plus_16b':
        partial(
            quant_espcn,
            upscale_factor=2,
            weight_bit_width=8,
            act_bit_width=8,
            acc_bit_width=16,
            weight_quant=CommonIntAccumulatorAwareZeroCenterWeightQuant)}

root_url = 'https://github.com/Xilinx/brevitas/releases/download/'

model_url = {
    'float_espcn_x2':
        f'{root_url}/super_res_r1/float_espcn_x2-2f85a454.pth',
    'quant_espcn_x2_w4a4_a2q_13b':
        f'{root_url}/super_res_r1/quant_espcn_x2_w4a4_a2q_13b-9fff234e.pth',
    'quant_espcn_x2_w4a4_a2q_32b':
        f'{root_url}/super_res_r1/quant_espcn_x2_w4a4_a2q_32b-8702a412.pth',
    'quant_espcn_x2_w4a4_base':
        f'{root_url}/super_res_r1/quant_espcn_x2_w4a4_base-80658e6d.pth',
    'quant_espcn_x2_w8a8_a2q_16b':
        f'{root_url}/super_res_r1/quant_espcn_x2_w8a8_a2q_16b-f9e1da66.pth',
    'quant_espcn_x2_w8a8_a2q_32b':
        f'{root_url}/super_res_r1/quant_espcn_x2_w8a8_a2q_32b-85470d9b.pth',
    'quant_espcn_x2_w8a8_base':
        f'{root_url}/super_res_r1/quant_espcn_x2_w8a8_base-f761e4a1.pth',
    'quant_espcn_x2_w4a4_a2q_plus_13b':
        f'{root_url}/super_res_r2/quant_espcn_x2_w4a4_a2q_plus_13b-6e6d55f0.pth',
    'quant_espcn_x2_w8a8_a2q_plus_16b':
        f'{root_url}/super_res_r2/quant_espcn_x2_w8a8_a2q_plus_16b-0ddf46f1.pth'}


def get_model_by_name(name: str, pretrained: bool = False) -> Union[FloatESPCN, QuantESPCN]:
    if name not in model_impl.keys():
        raise NotImplementedError(f"{name} does not exist.")
    model: nn.Module = model_impl[name]()
    if pretrained:
        if name not in model_impl:
            raise NotImplementedError(f"Error: {name} does not have a pre-trained checkpoint.")
        checkpoint = model_url[name]
        state_dict = hub.load_state_dict_from_url(checkpoint, progress=True, map_location='cpu')
        model.load_state_dict(state_dict, strict=True)
    return model



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from dependencies import this
from dependencies import value
from torch import nn

from brevitas.core.bit_width import BitWidthConst
from brevitas.core.bit_width import BitWidthStatefulConst
from brevitas.core.function_wrapper import Identity
from brevitas.core.function_wrapper import OverOutputChannelView
from brevitas.core.function_wrapper import RoundToZeroSte
from brevitas.core.function_wrapper import TensorClamp
from brevitas.core.function_wrapper import TensorClampSte
from brevitas.core.function_wrapper.ops_ste import CeilSte
from brevitas.core.function_wrapper.shape import StatsInputViewShapeImpl
from brevitas.core.quant import ClampedBinaryQuant
from brevitas.core.quant.int import DecoupledRescalingIntQuant
from brevitas.core.quant.int import DecoupledRescalingIntQuantWithInput
from brevitas.core.quant.int_base import DecoupledIntQuant
from brevitas.core.restrict_val import FloatRestrictValue
from brevitas.core.restrict_val import LogFloatRestrictValue
from brevitas.core.scaling import AccumulatorAwareParameterPreScaling
from brevitas.core.scaling import AccumulatorAwareZeroCenterParameterPreScaling
from brevitas.core.scaling import IntScaling
from brevitas.core.scaling import ParameterFromStatsFromParameterScaling
from brevitas.core.scaling import ParameterPreScalingWeightNorm
from brevitas.core.scaling import ParameterScaling
from brevitas.core.scaling import SCALAR_SHAPE
from brevitas.core.scaling import SCALING_STATS_REDUCE_DIM
from brevitas.core.scaling import StatsFromParameterScaling
from brevitas.core.stats import AbsMax
from brevitas.core.stats import AbsMaxL2
from brevitas.core.stats import AbsMinMax
from brevitas.core.stats import L1Norm
from brevitas.core.stats import L2Norm
from brevitas.core.stats import MSE
from brevitas.core.stats import NegativeMinOrZero
from brevitas.core.stats import NegativePercentileOrZero
from brevitas.core.stats.stats_op import HalfQuadraticOptimizerScale
from brevitas.core.stats.stats_op import HalfQuadraticOptimizerZeroPoint
from brevitas.core.utils import SingleArgStatelessBuffer
from brevitas.core.zero_point import ParameterFromRuntimeZeroPoint
from brevitas.core.zero_point import ParameterFromStatsFromParameterZeroPoint
from brevitas.core.zero_point import PreZeroCenterZeroPoint
from brevitas.core.zero_point import StatsFromParameterZeroPoint
from brevitas.core.zero_point import ZeroZeroPoint
from brevitas.inject import ExtendedInjector
from brevitas.inject.enum import BitWidthImplType
from brevitas.inject.enum import FloatToIntImplType
from brevitas.inject.enum import QuantType
from brevitas.inject.enum import RestrictValueType
from brevitas.inject.enum import ScalingImplType
from brevitas.inject.enum import ScalingPerOutputType
from brevitas.inject.enum import StatsOp
from brevitas.proxy import DecoupledWeightQuantProxyFromInjector
from brevitas.proxy import DecoupledWeightQuantWithInputProxyFromInjector
from brevitas.quant.solver.common import SolveStatsReduceDimFromEnum
from brevitas.quant.solver.parameter import SolveInputViewImpl
from brevitas.quant.solver.parameter import SolveParameterScalingShape
from brevitas.quant.solver.weight import SolveWeightScalingPerOutputChannelShapeFromModule
from brevitas.quant.solver.weight import SolveWeightScalingStatsInputDimsFromModule

__all__ = [
    'MaxStatsScaling',
    'MinMaxStatsScaling',
    'ParamFromRuntimePercentileScaling',
    'ParamFromRuntimePercentileIntervalScaling',
    'ParamFromRuntimeMinMaxScaling',
    'ParamMinMaxInitScaling',
    'IntQuant',
    'NarrowIntQuant',
    'UintQuant',
    'ShiftedMinUintQuant',
    'ShiftedParamFromPercentileUintQuant',
    'PerChannelFloatScaling8bit',
    'PerTensorFloatScaling8bit',
    'PerTensorPoTScaling8bit',
    'PerChannelPoTScaling8bit',
    'SignedBinaryClampedConst',
    'WeightPerTensorFloatDecoupledL2Param',
    'WeightPerChannelFloatDecoupled',
    'WeightNormPerChannelFloatDecoupled',
    'BatchQuantStatsScaling1d',
    'BatchQuantStatsScaling2d',
    'AccumulatorAwareWeightQuant',
    'AccumulatorAwareZeroCenterWeightQuant',
    'MSESymmetricScale',
    'MSEAsymmetricScale',
    'MSEWeightZeroPoint',
    'MSEActZeroPoint']


class MaxStatsScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.STATS
    scaling_stats_op = StatsOp.MAX
    scaling_min_val = 1e-10


class MinMaxStatsScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.STATS
    scaling_stats_op = StatsOp.MIN_MAX
    scaling_min_val = 1e-10


class ParamFromRuntimePercentileScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_op = StatsOp.PERCENTILE
    high_percentile_q = 99.999
    collect_stats_steps = 300
    scaling_min_val = 1e-10


class BatchQuantStatsScaling1d(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.AFFINE_STATS
    scaling_stats_op = StatsOp.MAX_AVE
    scaling_stats_momentum = None  # perform a true average rather than exp average
    affine_shift_scale = False  # max_ave statistics are scaled but not shifted
    scaling_stats_permute_dims = (1, 0, 2)  # assuming (N, C, F), put channel dimension first
    scaling_min_val = 1e-10


class BatchQuantStatsScaling2d(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.AFFINE_STATS
    scaling_stats_op = StatsOp.MAX_AVE
    scaling_stats_momentum = None  # perform a true average rather than exp average
    affine_shift_scale = False  # max_ave statistics are scaled but not shifted
    scaling_stats_permute_dims = (1, 0, 2, 3)  # assuming (N, C, H, W), put channel dimension first
    scaling_min_val = 1e-10


class ParamFromRuntimeMinMaxScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_op = StatsOp.MIN_MAX
    collect_stats_steps = 300
    scaling_min_val = 1e-10


class ParamFromRuntimePercentileIntervalScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_op = StatsOp.PERCENTILE_INTERVAL
    high_percentile_q = 99.999
    low_percentile_q = 0.001
    collect_stats_steps = 300
    scaling_min_val = 1e-10


class ParamMinMaxInitScaling(ExtendedInjector):
    """
    """
    scaling_impl_type = ScalingImplType.PARAMETER


class IntQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = True
    zero_point_impl = ZeroZeroPoint


class NarrowIntQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = True
    signed = True
    zero_point_impl = ZeroZeroPoint


class UintQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = False
    zero_point_impl = ZeroZeroPoint


class ShiftedMinUintQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = False
    quantize_zero_point = True
    zero_point_impl = StatsFromParameterZeroPoint
    zero_point_stats_impl = NegativeMinOrZero
    zero_point_shape = this.scaling_shape
    zero_point_stats_input_view_shape_impl = this.scaling_stats_input_view_shape_impl
    zero_point_stats_input_concat_dim = this.scaling_stats_input_concat_dim


class ShiftedParamFromPercentileUintQuant(ExtendedInjector):
    """
    """
    quant_type = QuantType.INT
    bit_width_impl_type = BitWidthImplType.CONST
    float_to_int_impl_type = FloatToIntImplType.ROUND
    narrow_range = False
    signed = False
    quantize_zero_point = True
    zero_point_impl = ParameterFromRuntimeZeroPoint
    zero_point_stats_impl = NegativePercentileOrZero
    low_percentile_q = 0.001
    zero_point_shape = this.scaling_shape
    zero_point_stats_input_view_shape_impl = this.scaling_stats_input_view_shape_impl


class PerChannelFloatScaling8bit(ExtendedInjector):
    """
    """
    scaling_per_output_type = ScalingPerOutputType.CHANNEL
    restrict_scaling_type = RestrictValueType.FP
    bit_width = 8


class PerTensorFloatScaling8bit(ExtendedInjector):
    """
    """
    scaling_per_output_type = ScalingPerOutputType.TENSOR
    restrict_scaling_type = RestrictValueType.FP
    bit_width = 8


class PerChannelPoTScaling8bit(ExtendedInjector):
    """
    """
    scaling_per_output_type = ScalingPerOutputType.CHANNEL
    restrict_scaling_type = RestrictValueType.FP
    bit_width = 8


class PerTensorPoTScaling8bit(ExtendedInjector):
    """
    """
    scaling_per_output_type = ScalingPerOutputType.TENSOR
    restrict_scaling_type = RestrictValueType.POWER_OF_TWO
    bit_width = 8
    restrict_value_float_to_int_impl = CeilSte


class SignedBinaryClampedConst(ExtendedInjector):
    tensor_quant = ClampedBinaryQuant
    scaling_impl_type = ScalingImplType.CONST
    restrict_scaling_type = RestrictValueType.FP
    float_to_int_impl_type = FloatToIntImplType.ROUND
    scaling_per_output_type = ScalingPerOutputType.TENSOR
    narrow_range = True
    signed = True


class PerTensorConstScaling2bit(ExtendedInjector):
    scaling_impl_type = ScalingImplType.CONST
    restrict_scaling_type = RestrictValueType.FP
    scaling_per_output_type = ScalingPerOutputType.TENSOR
    bit_width = 2


class WeightPerTensorFloatDecoupledL2Param(SolveWeightScalingStatsInputDimsFromModule):
    """
    Experimental narrow per-tensor signed int weight quantizer fragment with decoupled L2,inf
    normalization and learned scaling.
    """

    proxy_class = DecoupledWeightQuantProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuant
    decoupled_int_quant = DecoupledIntQuant
    tensor_clamp_impl = TensorClampSte
    pre_scaling_impl = StatsFromParameterScaling
    scaling_stats_impl = AbsMaxL2
    scaling_stats_input_view_shape_impl = OverOutputChannelView
    scaling_stats_input_concat_dim = 0
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    restrict_scaling_impl = FloatRestrictValue
    scaling_shape = SCALAR_SHAPE
    scaling_per_output_type = ScalingPerOutputType.TENSOR
    input_view_impl = Identity
    scaling_impl = ParameterFromStatsFromParameterScaling
    int_scaling_impl = IntScaling
    zero_point_impl = ZeroZeroPoint
    pre_zero_point_impl = ZeroZeroPoint
    bit_width_impl = BitWidthConst
    narrow_range = True
    signed = True


class WeightPerChannelFloatDecoupled(SolveStatsReduceDimFromEnum,
                                     SolveWeightScalingStatsInputDimsFromModule,
                                     SolveWeightScalingPerOutputChannelShapeFromModule,
                                     SolveParameterScalingShape,
                                     SolveInputViewImpl):
    """
    Experimental narrow per-channel signed int weight quantizer fragment with decoupled Linf
    normalization and learned scaling.
    """

    proxy_class = DecoupledWeightQuantProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuant
    decoupled_int_quant = DecoupledIntQuant
    tensor_clamp_impl = TensorClampSte
    pre_scaling_impl = StatsFromParameterScaling
    scaling_stats_impl = AbsMax
    restrict_scaling_impl = FloatRestrictValue
    scaling_impl = ParameterFromStatsFromParameterScaling
    int_scaling_impl = IntScaling
    zero_point_impl = ZeroZeroPoint
    pre_zero_point_impl = ZeroZeroPoint
    bit_width_impl = BitWidthConst
    narrow_range = True
    signed = True
    scaling_stats_input_view_shape_impl = OverOutputChannelView
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    scaling_per_output_type = ScalingPerOutputType.CHANNEL


class PerChannelL2Norm(ExtendedInjector):
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    normalize_stats_impl = L2Norm


class PerChannelL1Norm(ExtendedInjector):
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    normalize_stats_impl = L1Norm


class PerChannelPreNorm(ExtendedInjector):
    pre_scaling_impl = ParameterPreScalingWeightNorm
    scaling_stats_input_view_shape_impl = OverOutputChannelView
    scaling_impl = (this << 1).scaling_impl
    normalize_stats_impl = (this << 1).normalize_stats_impl
    tracked_parameter_list = (this << 1).tracked_parameter_list
    pre_scaling_shape = (this << 1).pre_scaling_shape
    permute_dims = (this << 1).permute_dims


class AccumulatorAwarePerChannelPreNorm(PerChannelPreNorm):

    pre_scaling_impl = AccumulatorAwareParameterPreScaling
    accumulator_bit_width = (this << 1).accumulator_bit_width
    accumulator_bit_width_impl = (this << 1).accumulator_bit_width_impl


class AccumulatorAwareZeroCenterPerChannelPreNorm(AccumulatorAwarePerChannelPreNorm):

    pre_scaling_impl = AccumulatorAwareZeroCenterParameterPreScaling
    pre_zero_point_impl = PreZeroCenterZeroPoint
    pre_zero_point_shape = this.pre_scaling_shape  # TODO: decouple zero_point from scaling
    pre_zero_point_stats_input_view_shape_impl = this.scaling_stats_input_view_shape_impl
    stats_reduce_dim = SCALING_STATS_REDUCE_DIM
    scaling_shape = (this << 1).scaling_shape


class SolvePostScaleGranularity(ExtendedInjector):

    @value
    def scaling_stats_input_view_shape_impl(scaling_per_output_type):
        if scaling_per_output_type == ScalingPerOutputType.TENSOR:
            return StatsInputViewShapeImpl.OVER_TENSOR
        elif scaling_per_output_type == ScalingPerOutputType.CHANNEL:
            return StatsInputViewShapeImpl.OVER_OUTPUT_CHANNELS

    @value
    def stats_reduce_dim(scaling_per_output_type):
        if scaling_per_output_type == ScalingPerOutputType.TENSOR:
            return None
        elif scaling_per_output_type == ScalingPerOutputType.CHANNEL:
            return SCALING_STATS_REDUCE_DIM


class WeightNormPerChannelFloatDecoupled(SolvePostScaleGranularity,
                                         SolveStatsReduceDimFromEnum,
                                         SolveWeightScalingStatsInputDimsFromModule,
                                         SolveWeightScalingPerOutputChannelShapeFromModule,
                                         SolveParameterScalingShape,
                                         SolveInputViewImpl):
    """Experimental narrow per-channel weight normalization-based signed integer quantizer
    based on `Quantized Neural Networks for Low-Precision Accumulation with Guaranteed
    Overflow Avoidance` by I. Colbert, A. Pappalardo, and J. Petri-Koenig.

    The formulation for weight normalization-based quantization is given below:
        `y = clip(round( (g / s) * (w / norm(w)) )) * s`

    The default quantizer uses the decoupled rescaling integer quantization arithmetic
    where the weight normalization calculation and parameterization are combined with the
    scaling factor to become the pre-clipping scaling factor (i.e., `pre_scale`) and the
    scaling factor is the post-clipping scaling factor (i.e., `post_scale`). For further
    details on the arithmetic, see `ParameterPreScalingWeightNorm`. For further details
    on the weight normalization-based quantization technique, see the referenced paper."""

    @value
    def scaling_init(scaling_init_impl, bit_width):
        scales = scaling_init_impl.parameter_list_stats() / (pow(2., bit_width - 1.) - 1.)
        return scales

    per_channel_pre_norm = PerChannelPreNorm

    proxy_class = DecoupledWeightQuantProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuant
    decoupled_int_quant = DecoupledIntQuant
    tensor_clamp_impl = TensorClamp
    scaling_impl = ParameterScaling
    scaling_init_impl = StatsFromParameterScaling
    restrict_scaling_impl = LogFloatRestrictValue
    scaling_stats_impl = AbsMax
    restrict_pre_scaling_impl = LogFloatRestrictValue
    normalize_stats_impl = PerChannelL2Norm.normalize_stats_impl
    scaling_per_output_type = ScalingPerOutputType.CHANNEL
    pre_scaling_shape = this.scaling_per_output_channel_shape
    int_scaling_impl = SingleArgStatelessBuffer(1.)
    zero_point_impl = ZeroZeroPoint
    pre_zero_point_impl = ZeroZeroPoint
    bit_width_impl = BitWidthConst
    narrow_range = True
    signed = True
    scaling_min_val = 1e-10
    pre_scaling_min_val = 1e-10

    @value
    def pre_scaling_impl():
        return this.per_channel_pre_norm.pre_scaling_impl


class AccumulatorAwareWeightQuant(WeightNormPerChannelFloatDecoupled):
    """Experimental accumulator-aware weight quantizer based on `Quantized Neural Networks
    for Low-Precision Accumulation with Guaranteed Overflow Avoidance` by I. Colbert,
    A. Pappalardo, and J. Petri-Koenig.

    The formulation is based on weight normalization-based quantization as given below:
        `y = clip(round( (g / s) * (w / norm(w)) )) * s`
    where `g` is the constrained such that the upper-bound on the l1-norm of the quantized
    weights satisfies the bounds derived in the referenced paper.

    The default quantizer uses input-aware decoupled rescaling integer quantization arithmetic
    where the constrained weight normalization calculation and parameterization are combined
    with the scaling factor to become the pre-clipping scaling factor (i.e., `pre_scale`) an
    the scaling factor is the post-clipping scaling factor (i.e., `post_scale`). For further
    details on the arithmetic, see `AccumulatorAwareParameterPreScalingWeightNorm`. For further
    details on accumulator-aware quantization (A2Q) technique, see the referenced paper."""

    proxy_class = DecoupledWeightQuantWithInputProxyFromInjector
    tensor_quant = DecoupledRescalingIntQuantWithInput
    per_channel_pre_norm = AccumulatorAwarePerChannelPreNorm
    normalize_stats_impl = PerChannelL1Norm.normalize_stats_impl  # required to align with derivations in paper
    float_to_int_impl = RoundToZeroSte  # required to ensure no upwards rounding violates constraints
    accumulator_bit_width = 32  # default maximum accumulator width is 32 bits

    @value
    def accumulator_bit_width_impl(accumulator_bit_width):
        return BitWidthStatefulConst(accumulator_bit_width)


class AccumulatorAwareZeroCenterWeightQuant(AccumulatorAwareWeightQuant):
    """Experimental zero-centered accumulator-aware weight quantized based on:
    `A2Q+: Improving Accumulator-Aware Weight Quantization`.

    When compared to A2Q, A2Q+ changes the following:
    (1) added zero-centering constraint on the weights (i.e., `PreZeroCenterZeroPoint`)
    (2) a more relaxed l1-norm bound that is derived in the referenced paper
    """
    per_channel_pre_norm = AccumulatorAwareZeroCenterPerChannelPreNorm

    @value
    def pre_zero_point_impl():
        return this.per_channel_pre_norm.pre_zero_point_impl


class MSESymmetricScaleSubInjector(ExtendedInjector):
    scaling_per_output = (this << 1).scaling_per_output
    proxy_module = (this << 1).proxy_module
    mse_init_op = AbsMax
    stats_impl = MSE
    stats_reduce_dim = (this << 1).stats_reduce_dim
    device = (this << 1).device
    type = (this << 1).type
    permute_dims = (this << 1).permute_dims
    inner_stats_input_view_shape_impl = (this << 1).inner_stats_input_view_shape_impl


class MSEAsymmetricScaleSubInjector(ExtendedInjector):
    scaling_per_output = (this << 1).scaling_per_output
    proxy_module = (this << 1).proxy_module
    mse_init_op = AbsMinMax
    stats_impl = MSE
    stats_reduce_dim = (this << 1).stats_reduce_dim
    device = (this << 1).device
    dtype = (this << 1).dtype
    permute_dims = (this << 1).permute_dims
    inner_stats_input_view_shape_impl = (this << 1).inner_stats_input_view_shape_impl


class MSEZeroPointSubInjector(ExtendedInjector):
    # zp is per channel when scaling is per channel
    scaling_per_output = (this << 1).scaling_per_output
    proxy_module = (this << 1).proxy_module
    mse_init_op = NegativeMinOrZero
    mse_search_method = 'grid'
    stats_impl = MSE
    stats_reduce_dim = (this << 1).stats_reduce_dim
    device = (this << 1).device
    dtype = (this << 1).dtype
    permute_dims = (this << 1).permute_dims
    inner_stats_input_view_shape_impl = (this << 1).inner_stats_input_view_shape_impl


class MSEAsymmetricScale(ExtendedInjector):
    """
    We leverage a sub-injector to avoid a name clash between scale and zero-point.
    """

    mse_scale = MSEAsymmetricScaleSubInjector
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_input_view_shape_impl = nn.Identity()

    @value
    def inner_stats_input_view_shape_impl(scaling_per_output):
        if scaling_per_output == ScalingPerOutputType.CHANNEL:
            return StatsInputViewShapeImpl.OVER_OUTPUT_CHANNELS
        elif scaling_per_output == ScalingPerOutputType.TENSOR:
            return StatsInputViewShapeImpl.OVER_TENSOR
        elif scaling_per_output == ScalingPerOutputType.GROUP:
            return StatsInputViewShapeImpl.OVER_SUBCHANNEL_BLOCK

    @value
    def scaling_stats_impl():
        return this.mse_scale.stats_impl


class MSESymmetricScale(ExtendedInjector):
    """
    We leverage a sub-injector to avoid a name clash between scale and zero-point.
    """

    mse_scale = MSESymmetricScaleSubInjector
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    scaling_stats_input_view_shape_impl = nn.Identity()

    @value
    def inner_stats_input_view_shape_impl(scaling_per_output):
        if scaling_per_output == ScalingPerOutputType.CHANNEL:
            return StatsInputViewShapeImpl.OVER_OUTPUT_CHANNELS
        elif scaling_per_output == ScalingPerOutputType.TENSOR:
            return StatsInputViewShapeImpl.OVER_TENSOR
        elif scaling_per_output == ScalingPerOutputType.GROUP:
            return StatsInputViewShapeImpl.OVER_SUBCHANNEL_BLOCK

    @value
    def scaling_stats_impl():
        return this.mse_scale.stats_impl


class MSEZeroPoint(ExtendedInjector):
    """
    We leverage a sub-injector to avoid a name clash between scale and zero-point.
    """

    mse_zero_point = MSEZeroPointSubInjector
    zero_point_stats_input_view_shape_impl = nn.Identity()

    @value
    def inner_stats_input_view_shape_impl(scaling_per_output):
        if scaling_per_output == ScalingPerOutputType.CHANNEL:
            return StatsInputViewShapeImpl.OVER_OUTPUT_CHANNELS
        elif scaling_per_output == ScalingPerOutputType.TENSOR:
            return StatsInputViewShapeImpl.OVER_TENSOR
        elif scaling_per_output == ScalingPerOutputType.GROUP:
            return StatsInputViewShapeImpl.OVER_SUBCHANNEL_BLOCK

    @value
    def zero_point_stats_impl():
        return this.mse_zero_point.stats_impl


class MSEWeightZeroPoint(MSEZeroPoint):
    zero_point_impl = ParameterFromStatsFromParameterZeroPoint


class MSEActZeroPoint(MSEZeroPoint):
    zero_point_impl = ParameterFromRuntimeZeroPoint


class HQOZeroPoint(ExtendedInjector):

    hqo_init_op_zp = NegativeMinOrZero
    inner_stats_input_view_shape_impl = this.zero_point_stats_input_view_shape_impl
    stats_impl_zp = HalfQuadraticOptimizerZeroPoint

    @value
    def zero_point_stats_impl():
        return this.stats_impl_zp


class HQOScale(ExtendedInjector):
    scaling_impl_type = ScalingImplType.PARAMETER_FROM_STATS
    inner_stats_input_view_shape_impl = this.scaling_stats_input_view_shape_impl
    stats_impl_scale = HalfQuadraticOptimizerScale

    @value
    def scaling_stats_impl():
        return this.stats_impl_scale


class HQOAsymmetricScale(HQOScale):
    hqo_init_op_scale = AbsMinMax


class HQOSymmetricScale(HQOScale):
    hqo_init_op_scale = AbsMax


class HQOActZeroPoint(HQOZeroPoint):
    zero_point_impl = ParameterFromRuntimeZeroPoint


class HQOWeightZeroPoint(HQOZeroPoint):
    zero_point_impl = ParameterFromStatsFromParameterZeroPoint
